/* $Workfile: HdlcExDemo1.c $       $Revision: 14 $      $Modtime: 1/03/07 1:46p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    HdlcExDemo1.c

    A simple example console application demonstrating the use of the
    software HDLC senders/receivers using OTX board's Burst device.

    The following boards are supported by this application:

			Thor-2(4,8)-PCI-Plus
      Sleipnir-1-PCI-Plus
      Gimle-16-PCI-Plus

    For Gimle-16-PCI-Plus, this demo configures HDLC receivers only on 16 Lines; 
    for other boards, it configures 1-8 Lines (base on board type) which can 
    be looped with an external loop cable.

    The direct sends and receives operate on a highway basis and are
    hosted by the Burst device.

    The programs configures 31 HDLC channels per Li (TS1 - TS31).

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

	  Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxBrd.h"

#include "OtxBurst.h"
#include "OtxBurstD.h"

#include "OtxHdlcD.h"
#include "OtxHdlcEx.h"
#include "OtxData.h"

/*******************************************************************************
 Defines
*******************************************************************************/

//#define USE_HDLCEX_BULK   // use the bulk io processing - ensures hdlc recv events are in chronological order


#define DEMO_DEFAULT_LOG_FILE   "HdlcExDemo1.log"
#define DEMO_DEFAULT_BOARD_NO   0
#define DEMO_DEFAULT_RX_THRESHOLD 500     // in mV
#define DEMO_BURST_BUF_SIZE     0x8000    // 32k

#define DEMO_TASK_REF           1L        // Task Reference for a non-blocking call
#define DEMO_HDLC_SENDER        1         // Comment this line to remove the HDLC Sender Logical devices from this application 
#define DEMO_MAX_MULTIPLE_FRAME_READ 128
#define DEMO_SS7_SIM_FRAMES     1000

#define DEMO_MAX_DEV            512

/*******************************************************************************
 Structures
*******************************************************************************/
typedef struct DemoEnvT {
  // Common 
  char* pszInputStr;
  OTX_BOOL bExitOnError;
  char cwd[OTX_MAX_PATH];
  char logFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int coLi;
  int coHighways;
  int coHdlcEx;
  OtxT1E1LiModeE eT1E1Mode;    // T1 or E1 mode
  int boardNo;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1E1TermModeE       eTermMode;
  OTX_BOOL               bMonitorMode;
  OTX_BOOL               bAmplifier;
  OTX_INT32              nRxThreshold;
  OTX_BOOL               bDriveTxPins;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OtxBrdClockSourceE     eClkSource;
  OTX_BOOL bVerboseLogging;

  OTX_BOOL bDevEnableList;
  OTX_INT32 sqDevEnableList[128];
  OTX_INT32 coDevEnableList;

  // HdlcEx related
  int coHdlcPerLi;
  OTX_BOOL bSs7FisuFiltering;
  OTX_BOOL bHdlcBitInversion;
  OTX_BOOL bSuperChannel;
  OTX_BOOL b56kbpsChannel;
  OTX_UINT32 nCapacity;
  int nTimeSlots;
  int nStartTS;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS* pE1CfgData, OUT OtxT1ConfigOptionsS* pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS* pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char* formatStr, ...);
static void DemoLogf2(char* formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char* argv[], DemoEnvT* envP);
static void DemoCreateLogicalDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintHdlcEvent(OtxEventDataS* pEventData);
static void DemoPrintEvent(OtxEventDataS* pEventData);
static void DemoDumpFalcRegisters(void);

void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_UINT32 g_coTimeSlotMask = 32;

OTX_EVENT g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE g_hEventQueue = 0;
FILE* g_pDemoLogFile = NULL;
FILE* g_pBurstSendFile = NULL;
FILE* g_pBurstRecvFile = NULL;

DemoEnvT g_demoEnv;
OTX_BOOL g_bStatsThreadActive = OTX_FALSE;

OTX_INT32 g_cbHdlcFrame0 = 10;
OTX_INT32 g_cbHdlcFrame1 = 10;
OTX_INT32 g_coHdlcSenderFrames = 1;
OTX_BOOL g_bHdlcVerbose = OTX_TRUE;
OTX_BOOL g_bHdlcCheckRxFrames = OTX_TRUE;
OTX_UINT32 g_coHdlcRxBytes[DEMO_MAX_DEV];
OTX_UINT32 g_coHdlcRxFrames[DEMO_MAX_DEV];
OTX_UINT32 g_coHdlcBadRxFrames[DEMO_MAX_DEV];
OTX_UINT8 g_sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
OTX_INT32 g_nSleepTimeBetweenHdlcFrames = 0;
OtxHdlcSS7FisuS g_tFisu0;
OtxHdlcSS7FisuS g_tFisu1;
OtxHdlcSS7ShortLssuS g_tLssu;

OTX_INT32 g_nSs7SimulationErrors[DEMO_MAX_DEV] = {
  0
};
OTX_BOOL g_bSs7SimulationActive[DEMO_MAX_DEV] = {
  0
};
OTX_INT32 g_nSs7SimulationFramesReceived[DEMO_MAX_DEV] = {
  0
};
OTX_UINT8 g_nPrevBsn[DEMO_MAX_DEV] = {
  0x7F
};

OTX_UINT32 g_BsnErrors = 0;
OTX_BOOL g_BsnCheck = OTX_FALSE;

OsaMutexS g_demoLogfCritSec;

struct {
  OTX_HANDLE hBoardController;
  OTX_HANDLE hLi[16];
  OTX_HANDLE hQuadLi[4];
  OTX_HANDLE hBurst;
} g_hPhysDevs;

struct {
  OTX_HANDLE hLiIn[16];
  OTX_HANDLE hLiOut[16];
} g_hPipes;

#ifdef DEMO_HDLC_SENDER
struct {
  HDLCX_HANDLE hHandle;
  OTX_INT32 nLi;
  OTX_UINT32 nTs;
  OTX_UINT32 nBurst;
} g_sqHdlcSender[DEMO_MAX_DEV];
#endif 

struct {
  HDLCX_HANDLE hHandle;
  OTX_INT32 nLi;
  OTX_UINT32 nTs;
  OTX_UINT32 nBurst;
} g_sqHdlcReceiver[DEMO_MAX_DEV];


// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32 nSourceId;
  OTX_BOOL bActive;
  FILE* pTxFile;
  OTX_INT32 nTsToSend;											// Default is -1 (All TS)
  OTX_BOOL bFileLoop;
  HDLCX_HANDLE hDev[DEMO_MAX_DEV];          // related HdlcEx devices
  OTX_UINT32 nDevs;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32 nSourceId;
  OTX_BOOL bActive;
  FILE* pRxFile;
  OTX_INT32 nTsToRecord;										// Default is -1 (All TS)
  HDLCX_HANDLE hDev[DEMO_MAX_DEV];
  OTX_UINT32 nDevs;
};

struct {
  struct sBurstSender mBurstSender[4];			// one for each possible highway hosted by the Burst device
  struct sBurstReceiver mBurstReceiver[4];  // one for each possible highway hosted by the Burst device
} g_hLogicDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************

  @func
    Initalize the global data for the Hdlc Sender and Receiver devices

  @comm

*******************************************************************************/
void DemoInitializeGlobalData(void)
{
  OTX_INT32 i, j, n, TxLi, RxLi, nTs;
  OTX_UINT32 nDivisor;

  OTX_UINT32 LiTable[16] = {
    // Tx 0, 1, 2, 3, 4, 5, 6, 7
    //      1, 0, 3, 2, 5, 4, 7, 6    // Cross-looped     
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13 ,14 ,15   // To itself     

  };

  if (g_demoEnv.bSuperChannel) {
    g_demoEnv.coHdlcPerLi = 1;
  }   // Only one superpipe per Li

  g_demoEnv.coHdlcEx = g_demoEnv.coLi * g_demoEnv.coHdlcPerLi;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
   case OTX_DEVICE_GIMLE_16_PCI_PLUS:
     g_coTimeSlotMask = 128; 
     break;
   case OTX_DEVICE_SLEIPNIR_1_PCI:
     g_coTimeSlotMask = 32;
     break;
   case OTX_DEVICE_THOR_8_PCI_PLUS:
     g_coTimeSlotMask = 128;
     break;
   case OTX_DEVICE_THOR_4_PCI_PLUS:
     g_coTimeSlotMask = 128;
     break;
   case OTX_DEVICE_THOR_2_PCI_PLUS:
     g_coTimeSlotMask = 32;
     break;
   default:
     DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
     OsaConsoleExit();
     OtxHdlcExReleaseLib();
     exit(1);
     break;
  }

  if ((OTX_UINT32) g_demoEnv.coHdlcEx >= g_demoEnv.coHighways * g_coTimeSlotMask) {
    DemoLogf("Error: Too big number of devices. Failed to map timeslots.\n");
    DemoLogf("Error: Could not connect to the OTX HdlcEx library.\n");
    OsaConsoleExit();
    OtxHdlcExReleaseLib();

    exit(1);
  }


  nDivisor = g_coTimeSlotMask / 32; // Number of E1 spans in one Burst highway

  if (g_demoEnv.coHighways * nDivisor == 1) {
    // One channel board. Loop it to itself
    LiTable[0] = 0;
  }

  n = 0;
  DemoLogf("Preparing timeslot mapping (used by burst device)\n");

  for (i = 0; i < g_demoEnv.coLi; i++) {
    TxLi = i % (g_demoEnv.coHighways * nDivisor); 
    RxLi = LiTable[TxLi] % (g_demoEnv.coHighways * nDivisor);
  
    for (j = 0; j < g_demoEnv.coHdlcPerLi; j++,n++) {
      // Timeslot in E1 span is the same both for Tx and Rx
      nTs = g_demoEnv.nStartTS + j; 
      if (nTs > 31) {
        continue;
      }

#ifdef DEMO_HDLC_SENDER
      g_sqHdlcSender[n].nLi = TxLi;
#endif
      g_sqHdlcReceiver[n].nLi = RxLi;

      // The difference is its offset in particular burst highways 
#ifdef DEMO_HDLC_SENDER
      g_sqHdlcSender[n].nTs = nTs + (TxLi % nDivisor) * 32; // Add offset of span among burst device hw
#endif
      g_sqHdlcReceiver[n].nTs = nTs + (RxLi % nDivisor) * 32;

#ifdef DEMO_HDLC_SENDER
      g_sqHdlcSender[n].nBurst = TxLi / nDivisor;
#endif
      g_sqHdlcReceiver[n].nBurst = RxLi / nDivisor;


      if (g_demoEnv.bVerboseLogging) {
        // Print mapping for debug purposes
        // It should be no cross-referenced timeslot in sender - otherwise data will be corrupted
        if(!g_demoEnv.bSuperChannel){

#ifdef DEMO_HDLC_SENDER
           DemoLogf("HdlcSender[%d]:   nLi=%d, nTs=%d, nBurst=%d\n",n, g_sqHdlcSender[n].nLi, g_sqHdlcSender[n].nTs, g_sqHdlcSender[n].nBurst);
#endif
           DemoLogf("HdlcReceiver[%d]: nLi=%d, nTs=%d, nBurst=%d\n",n, g_sqHdlcReceiver[n].nLi, g_sqHdlcReceiver[n].nTs, g_sqHdlcReceiver[n].nBurst);
        } else {

#ifdef DEMO_HDLC_SENDER
           DemoLogf("HdlcSender[%d]:   nLi=%d, nTs=(%d..%d), nBurst=%d\n",n, g_sqHdlcSender[n].nLi, g_sqHdlcSender[n].nTs, g_sqHdlcSender[n].nTs+g_demoEnv.nTimeSlots-1, g_sqHdlcSender[n].nBurst);
#endif
           DemoLogf("HdlcReceiver[%d]: nLi=%d, nTs=(%d..%d), nBurst=%d\n",n, g_sqHdlcReceiver[n].nLi, g_sqHdlcReceiver[n].nTs, g_sqHdlcReceiver[n].nTs+g_demoEnv.nTimeSlots-1, g_sqHdlcReceiver[n].nBurst);
        }
      }
    }
  }

  DemoLogf("Timeslot mapping complete\n");
}


/*******************************************************************************

  @func

*******************************************************************************/
void DemoPrintSignalSettings(void)
{
  // Signal settings
  char *pszTermMode;

  switch (g_demoEnv.eTermMode) {
  case OTX_T1E1_TERM_120_OHM:
    pszTermMode = "120  ";
    break;
  case OTX_T1E1_TERM_100_OHM:
    pszTermMode = "100  ";
    break;
  case OTX_T1E1_TERM_75_OHM:
    pszTermMode = "75U  ";
    break;
  case OTX_T1E1_TERM_HIGHZ:
    pszTermMode = "HIGHZ";
    break;
  case OTX_T1E1_TERM_75_OHM_BALANCED:
    pszTermMode = "75B  ";
    break;
  default:
    pszTermMode = "???  ";
    break;
  }

  DemoLogf("RxEQ=%s, TermMode=%s, MonMode=%s, Amp=%s, RxThold=%.4d, TxPins=%s\n",
    g_demoEnv.bReceiveEqualizer ? "ON " : "OFF",
    pszTermMode,
    g_demoEnv.bMonitorMode ? "ON " : "OFF",
    g_demoEnv.bAmplifier ? "ON " : "OFF",
    g_demoEnv.nRxThreshold,
    g_demoEnv.bDriveTxPins ? "DRIVE" :"HIGHZ"
  );

}




void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

  @comm

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(2048, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);
}

  
void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-8 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
   case OTX_DEVICE_GIMLE_16_PCI_PLUS:
     DemoLogf("Opening Gimle-16-PCI-Plus board\n");
     break;
   case OTX_DEVICE_SLEIPNIR_1_PCI:
     DemoLogf("Opening Sleipnir board\n");
     break;
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      DemoLogf("Opening Thor-2-PCI Plus board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      DemoLogf("Opening Thor-4-PCI Plus board\n");
      break;
   case OTX_DEVICE_THOR_8_PCI_PLUS:
     DemoLogf("Opening Thor-8-PCI Plus board\n");
     break;
   default:
     DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
     OsaConsoleExit();
     exit(1);
     break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController); DemoCheckRc(g_hPhysDevs.hBoardController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the 2 individual T1/E1 line interface members within the Quad device
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hBoardController, result);
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
    for (i = 0; i < 1; i++) {
      // Get handles to the one quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i*4, g_hEventQueue, &g_hPhysDevs.hLi[j + i*4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the two quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i * 4, g_hEventQueue, &g_hPhysDevs.hLi[j + i * 4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    // Open the Line interfaces
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hBoardController, result);
    }

    // we'll source the Falc from the Burst device, and source the burst device from the FALC
    //  Falc <-- Burst Send (P2S)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(g_hPhysDevs.hBoardController, result);
    //  Burst Recv (S2P) <-- Falc
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBoardController, OTX_ATTR_BRD_S2P_SOURCED, OTX_BRD_S2P_SOURCED_FALC); DemoCheckRc(g_hPhysDevs.hBoardController, result);

	} else if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) {
    for (i = 0; i < 4; i++) {
      // Get handles to the four quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i * 4, g_hEventQueue, &g_hPhysDevs.hLi[j + i * 4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }
	}

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hBoardController, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
}

  
/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
static void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32 i;

  DemoPrintSignalSettings();

  // Set default configuration data
  DemoInitT1E1ConfigData(&(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));


  for (i = 0; i < g_demoEnv.coLi; i++) {
     
    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], g_demoEnv.eTermMode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
		    
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1
      // Configure T1 Transceivers
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      
      // Set them in T1 mode
      result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1
      // Configure E1 Transceivers
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      
      // Set them in E1 mode
      result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    
    // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
    DemoLogf("Li[%d]: Setting XmitPinControl to %s\n", i, g_demoEnv.bDriveTxPins ? "DRIVE" : "TRISTATE");
    result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], g_demoEnv.bDriveTxPins); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Set receive threshold
    DemoLogf("Li[%d]: Setting RX Input Threshold to %d\n", i, g_demoEnv.nRxThreshold);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    { // Read back the threshold
      OTX_UINT32 nRxThreshold;
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      DemoLogf("Li[%d]: RX Input Threshold = %d\n", i, nRxThreshold);
    }

    // Set Monitor (-20dB) mode
    DemoLogf("Li[%d]: Setting Monitor mode to %s\n", i, g_demoEnv.bMonitorMode ? "ON" : "OFF");
    result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[i], g_demoEnv.bMonitorMode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Turning front end Amplifier ON/OFF      
    DemoLogf("Li[%d]: Setting Front End Amplifier to %s\n", i, g_demoEnv.bAmplifier ? "ON" : "OFF");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bAmplifier); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    
  }



}


void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

  @comm

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs.hBoardController\n");
  result = OtxDrvEnable(g_hPhysDevs.hBoardController, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
    for (i = 0; i < 1; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }
  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }
  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS) {
    for (i = 0; i < 4; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Enable Physical Devices, Li's
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(void)
{
  OTX_INT32 nLi;
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];

  DemoLogf("Creating pipe for Burst Sender/Receiver\n");

	// Create the Pipes (one for each direction) between the Burst Device and the LIs
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    OTX_INT32 nTs;

    // we must define the highway to use [0-3] 
    iHighwayNo = nLi / (g_demoEnv.coLi / g_demoEnv.coHighways);

    // Create pipes (one for each direction) Use all timeslots except TS0.
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask) * sizeof(sqBurstTimeSlotMask[0]));

    for (nTs = 0; nTs < 32; nTs++) {
      sqTimeSlotMask[nTs] = 0xFF;

      // for this demo, we'll lay out the burst time slot mask as follows:
      //    [ 0..31]  - LI0:TS0-TS31 and LI4:TS0-TS31
      // OTX_DEVICE_THOR_8_PCI_PLUS, OTX_DEVICE_GIMLE_16_PCI_PLUS only:
      //    [32..63]  - LI1:TS0-TS31 and LI5:TS0-TS31
      //    [64..95]  - LI2:TS0-TS31 and LI6:TS0-TS31
      //    [96..127] - LI3:TS0-TS31 and LI7:TS0-TS31
      //
      // upon connecting the pipe (see below), we have to pass the highway
      // number (0-3 - one highway per each QuadFalc (128TS per highway)
      // as well as the size of the highway (128 TS)
      //
      // this is just a simple algorithm to accomplish the above
      sqBurstTimeSlotMask[nTs + ((nLi % (g_demoEnv.coLi / g_demoEnv.coHighways)) * 32)] = 0xFF;
    }


    // --------------------------------------------------------------------
    // Create an incoming and outgoing Pipe representing the external T1/E1
    // --------------------------------------------------------------------
    result = OtxDrvCreatePipe(64 * 32, &g_hPipes.hLiIn[nLi]);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);
    result = OtxDrvCreatePipe(64 * 32, &g_hPipes.hLiOut[nLi]);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

#ifdef DEMO_HDLC_SENDER
    // -------------------------
    // Connect pipe: Burst -> Li
    // -------------------------
		// 14 Jan 05: modified coMaskSize to pass ConnectPipe validation performed by driver								 
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[nLi], OTX_PIPE_INPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_coTimeSlotMask);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);

    result = OtxDrvConnectPipe(g_hPipes.hLiIn[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi], -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);
#endif

    // -------------------------
    // Connect pipe: Li -> Burst
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_INPUT, g_hPhysDevs.hLi[nLi], -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

	  // 14 Jan 05: modified coMaskSize to pass ConnectPipe validation performed by driver								 
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_coTimeSlotMask );
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

	}
}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

  @comm

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS* pE1CfgData, 
  OUT OtxT1ConfigOptionsS* pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits = 0x3;
  pE1CfgData->bfSaBits = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe = OTX_TRUE;
  pE1CfgData->bAutoResync = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion = OTX_TRUE;
  pE1CfgData->bAutoRemoteAlarm = OTX_FALSE;
  pE1CfgData->nLosSensitivity = 10;
  pE1CfgData->nLosRecovery = 21;
  pE1CfgData->nLineLength = 10;
  pE1CfgData->bHighTxPower = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion = OTX_FALSE;
  pT1CfgData->eLfaSensitivity = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm = OTX_FALSE;
  pT1CfgData->nLosSensitivity = 10;
  pT1CfgData->nLosRecovery = 21;
  pT1CfgData->nLineLength = 10;
  pT1CfgData->bHighTxPower = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels = 0x00;
}


/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32 i, n;

  for (i = 0; i < g_demoEnv.coHdlcEx; i++) {
#ifdef DEMO_HDLC_SENDER
    result = OtxHdlcExCreate(OTX_LDEVICE_HDLC_SENDER, g_hEventQueue, i, &g_sqHdlcSender[i].hHandle);  DemoCheckRc(g_sqHdlcSender[i].hHandle, result);
#endif

    // Note that the HDLC Receiver is not using an event queue. Instead we register a call back function for the events.
    result = OtxHdlcExCreate(OTX_LDEVICE_HDLC_RECEIVER, OTX_INVALID_HANDLE_VALUE, i, &g_sqHdlcReceiver[i].hHandle);  DemoCheckRc(g_sqHdlcReceiver[i].hHandle, result);

    // Register the callback function for the HDLC receiver events
    result = OtxHdlcExSetCallback(g_sqHdlcReceiver[i].hHandle, &DemoPrintHdlcEvent); DemoCheckRc(g_sqHdlcReceiver[i].hHandle, result);

    //Fill burst device mapping table
#ifdef DEMO_HDLC_SENDER
    n = g_sqHdlcSender[i].nBurst;    g_hLogicDevs.mBurstSender[n].hDev[g_hLogicDevs.mBurstSender[n].nDevs++] = g_sqHdlcSender[i].hHandle;
#endif
    n = g_sqHdlcReceiver[i].nBurst;  g_hLogicDevs.mBurstReceiver[n].hDev[g_hLogicDevs.mBurstReceiver[n].nDevs++] = g_sqHdlcReceiver[i].hHandle;
  }
}



/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectLogicalDevices(void)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[128];

  if (g_demoEnv.bSuperChannel) {
    // x*64kbps HDLC capacity (TS1-TS31 E1 only)

    OTX_INT32 j;
    OTX_INT32 noOfTimeSlots = g_demoEnv.nTimeSlots;

    DemoLogf("Configuring one %dkbps HDLC channel per Li\n", 64 * g_demoEnv.nTimeSlots);

    for (i = 0; i < g_demoEnv.coHdlcEx; i++) {

#ifdef DEMO_HDLC_SENDER
      // Timeslot on the E1 span
      memset(sqTimeSlotMask, 0x00, g_coTimeSlotMask);
      for (j = 0; j < noOfTimeSlots; j++) {
        //        if(j==9 || j==10) continue;
        sqTimeSlotMask[g_sqHdlcSender[i].nTs + j] = 0xFF;
      }
      result = OtxHdlcExSetTimeSlotMask(g_sqHdlcSender[i].hHandle, g_coTimeSlotMask, sqTimeSlotMask);      DemoCheckRc(g_sqHdlcSender[i].hHandle, result);
#endif

      memset(sqTimeSlotMask, 0x00, g_coTimeSlotMask);
      for (j = 0; j < noOfTimeSlots; j++) {
        //        if(j==9 || j==10) continue;
        sqTimeSlotMask[g_sqHdlcReceiver[i].nTs + j] = 0xFF;
      }
      result = OtxHdlcExSetTimeSlotMask(g_sqHdlcReceiver[i].hHandle, g_coTimeSlotMask, sqTimeSlotMask);      DemoCheckRc(g_sqHdlcReceiver[i].hHandle, result);
    }
  } else {
	  OTX_INT32 nCapacity = g_demoEnv.nCapacity;
  	OTX_UINT8 mask = 0;

	  /*
      // 64kbps or 56kbps HDLC capacity
      if (g_demoEnv.b56kbpsChannel) {
        DemoLogf("56kbps HDLC channel per Li is not supported by this demo. Exiting..\n");
        return; // Not Supported.
      } else {
        DemoLogf("Configuring one 64kbps HDLC channel per Li\n");
        nCapacity = 64;
      }

	  */

    for(i=0; i< nCapacity/8; i++) mask = (mask << 1) | 1;


    for (i = 0; i < g_demoEnv.coHdlcEx; i++) {


#ifdef DEMO_HDLC_SENDER
      memset(sqTimeSlotMask, 0x00, g_coTimeSlotMask);
      sqTimeSlotMask[g_sqHdlcSender[i].nTs] = mask;
      result = OtxHdlcExSetTimeSlotMask(g_sqHdlcSender[i].hHandle, g_coTimeSlotMask, sqTimeSlotMask);      
      DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
#endif

      memset(sqTimeSlotMask, 0x00, g_coTimeSlotMask);
      sqTimeSlotMask[g_sqHdlcReceiver[i].nTs] = mask;
      result = OtxHdlcExSetTimeSlotMask(g_sqHdlcReceiver[i].hHandle, g_coTimeSlotMask, sqTimeSlotMask);      
      DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    }
  } // g_demoEnv.bSuperChannel
}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

  @comm

*******************************************************************************/
void DemoInitializeLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coHdlcEx; i++) {
#ifdef DEMO_HDLC_SENDER

    // Initialize the Hdlc Sender 
    result = OtxHdlcExSenderInit(
      g_sqHdlcSender[i].hHandle,  // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS, 
       /*OTX_HDLC_FFT_ALL_ONES,*/    // Fill character(s) to be send between HDLC frames
      1,                             // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,                // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      g_demoEnv.bHdlcBitInversion,   // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                      // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
      OTX_HDLC_TR_NA                 // HDLC Transmission Rate for V.110 and X.30 modes
    ); 
    DemoCheckRc(g_sqHdlcSender[i].hHandle, result);          
#endif

    // Initialize the Hdlc Receiver 
    result = OtxHdlcExReceiverInit(
      g_sqHdlcReceiver[i].hHandle, // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS
      /*OTX_HDLC_FFT_ALL_ONES*/,      // Fill character(s) to be send between HDLC frames
      1,                              // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,                 // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      g_demoEnv.bHdlcBitInversion,    // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                       // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
      OTX_HDLC_TR_NA,                 // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_TRUE                        // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
    ); 
    DemoCheckRc(g_sqHdlcReceiver[i].hHandle, result);          

    if (g_demoEnv.bSs7FisuFiltering) {
      //if (i == 1) continue;
      result = OtxHdlcExSS7SetFilter(g_sqHdlcReceiver[i].hHandle, OTX_HDLC_SS7_FILTER_FISU | OTX_HDLC_SS7_FILTER_SHORT_LSSU | OTX_HDLC_SS7_FILTER_LONG_LSSU); 
      DemoCheckRc(g_sqHdlcReceiver[i].hHandle, result);
    }
  }
}
  

/*******************************************************************************

  @func
    Enable Configured Logical Devices

  @comm

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32 i;

#ifdef DEMO_HDLC_SENDER
  DemoLogf("Enabling %d HDLC senders\n", g_demoEnv.coHdlcEx);
  for (i = 0; i < g_demoEnv.coHdlcEx; i++) {
    result = OtxHdlcExEnable(g_sqHdlcSender[i].hHandle); DemoCheckRc(g_sqHdlcSender[i].hHandle, result);
  }
#endif

  if (g_demoEnv.bDevEnableList) {
    // Enable specific receivers only as specified in the sqDevEnableList
    for (i = 0; i < g_demoEnv.coDevEnableList; i++) {
      DemoLogf("Enabling HDLC receiver [%d]\n", g_demoEnv.sqDevEnableList[i]);
      result = OtxHdlcExEnable(g_sqHdlcReceiver[g_demoEnv.sqDevEnableList[i]].hHandle); DemoCheckRc(g_sqHdlcReceiver[g_demoEnv.sqDevEnableList[i]].hHandle, result);
    }
  } else {
    // Enable all receivers
    DemoLogf("Enabling %d HDLC receivers\n", g_demoEnv.coHdlcEx);
    for (i = 0; i < g_demoEnv.coHdlcEx; i++) {
      result = OtxHdlcExEnable(g_sqHdlcReceiver[i].hHandle); DemoCheckRc(g_sqHdlcReceiver[i].hHandle, result);
    }
  }
}
  

#ifdef DEMO_HDLC_SENDER
/*******************************************************************************

  @func

  Transmits one or more HDLC frames

  @comm

*******************************************************************************/
void* DemoHdlcTransmitFunction(void* lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 busyCount = 0;
  clock_t startTime, stopTime;
  OTX_UINT32 nDev;

  nDev = (OTX_INT32) lpParameter;

#if 1 
  for (i = 0; i < g_cbHdlcFrame0; i++) {
    g_sqHdlcFrame[i] = (OTX_UINT8) i;
  }
#else // Testing zero-bit insertion of the last bit of the frame 
  g_cbHdlcFrame0 = 3;
  g_sqHdlcFrame[0] = 0xEC;
  g_sqHdlcFrame[1] = 0x49;
  g_sqHdlcFrame[2] = 0x00;
#endif

  DemoLogf("Sending %d HDLC frame(s) from HDLC Sender #%d (len=%d)\n", g_coHdlcSenderFrames, nDev, g_cbHdlcFrame0);

  // Send HDLC Frame(s)          
  startTime = clock();                 
  for (i = 0; i < g_coHdlcSenderFrames; i++) {
                                                   
    result = OtxHdlcExSendData(g_sqHdlcSender[nDev].hHandle, g_sqHdlcFrame, g_cbHdlcFrame0, DEMO_TASK_REF, OTX_TIME_INFINITY); 

    if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
      // Driver IO Request buffer is full. We need to retransmit this frame.
      i--;
      busyCount++;
    } else {
      DemoCheckRc(g_sqHdlcSender[nDev].hHandle, result);
      OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
      //nHdlcTaskRef++;
    }
  }
  stopTime = clock();

  printf("[%d]Done (BusyCount=%d, time=%2.1f seconds).\n", nDev, busyCount, (double) (stopTime - startTime) / CLOCKS_PER_SEC);

  return 0;
}
#endif


#ifdef DEMO_HDLC_SENDER
/*******************************************************************************

  @func

  Transmits one or more MSU frames

  @comm

*******************************************************************************/
void* DemoMsuTransmitFunction(
  void* lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 busyCount = 0;
  clock_t startTime, stopTime;
  OtxHdlcSS7FisuS msu;
  OtxHdlcSS7FisuS* pFisu;
  OTX_UINT8 sqHdlcFrame[300];
  OTX_INT32 cbHdlcFrame;
  OTX_UINT32 nDev;

  nDev = *((OTX_INT32 *) lpParameter);


  // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
  if (nDev == 0) {
    pFisu = &g_tFisu0;
    cbHdlcFrame = g_cbHdlcFrame0;
  } else {
    pFisu = &g_tFisu1;
    cbHdlcFrame = g_cbHdlcFrame1;
  }

  memcpy(&msu, pFisu, OTX_HDLC_SS7_CB_FISU);  // Build the first 3 bytes of the MSU
  msu.li = (OTX_UINT8) (cbHdlcFrame - OTX_HDLC_SS7_CB_FISU);  // Set the Length Indicator
  memcpy(sqHdlcFrame, &msu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
  for (i = OTX_HDLC_SS7_CB_FISU; i < cbHdlcFrame; i++) {
    // Add the payload the HDLC frame buffer
    sqHdlcFrame[i] = (OTX_UINT8) i - OTX_HDLC_SS7_CB_FISU;
  }

  DemoLogf("Sending %d MSU frame(s) (len=%d) from HDLC Sender #%d  Li#%d\n", g_coHdlcSenderFrames, cbHdlcFrame, nDev, g_sqHdlcSender[nDev].nLi);

  // Send MSU Frame(s)
  startTime = clock();
  for (i = 0; i < g_coHdlcSenderFrames; i++) {
    result = OtxHdlcExSS7SendData(g_sqHdlcSender[nDev].hHandle, sqHdlcFrame, cbHdlcFrame, pFisu, DEMO_TASK_REF, OTX_TIME_INFINITY); 
    if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
      // Driver IO Request buffer is full. We need to retransmit this frame.
      i--;
      busyCount++;
    } else {
      DemoCheckRc(g_sqHdlcSender[nDev].hHandle, result);
      OsaThreadSleep(g_nSleepTimeBetweenHdlcFrames);
    }
  }
  stopTime = clock();

  DemoLogf("[%d]Done (BusyCount=%d, time=%2.1f seconds).\n", nDev, busyCount, (double) (stopTime - startTime) / CLOCKS_PER_SEC);

  return 0;
}
#endif 


#ifdef DEMO_HDLC_SENDER
/*******************************************************************************

  @func

  Transmits one or more MSU frames

  @comm

*******************************************************************************/
void* DemoSs7SimulationFunction(
  void* lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 busyCount = 0;
  clock_t startTime, stopTime;
  OtxHdlcSS7FisuS msu;
  OtxHdlcSS7FisuS fisu;
  OTX_UINT8 sqHdlcFrame[200];
  OTX_INT32 cbHdlcFrame;
  OTX_UINT32 nDev;
  OTX_INT32 coHdlcSenderFrames = DEMO_SS7_SIM_FRAMES;
  OTX_INT32 coMsusSent = 0;
  OTX_INT32 coFisusSent = 0;

  nDev = *((OTX_INT32 *) lpParameter);

  fisu.bib = 1;
  fisu.fib = 1;
  fisu.fsn = 0x7F;
  fisu.bsn = 0x7F;
  fisu.li = OTX_HDLC_SS7_CB_FISU;
  fisu.reserved = 0;

  g_bSs7SimulationActive[nDev] = OTX_TRUE;

  memcpy(sqHdlcFrame, &fisu, OTX_HDLC_SS7_CB_FISU);  
  cbHdlcFrame = OTX_HDLC_SS7_CB_FISU;

  DemoLogf("Sending a mix of MSU and FISU frames from HDLC Sender #%d  (Li#%d)\n", nDev, g_sqHdlcSender[nDev].nLi);

  // Send MSU Frame(s)
  startTime = clock();
  while (coMsusSent < coHdlcSenderFrames) {
    result = OtxHdlcExSS7SendData(g_sqHdlcSender[nDev].hHandle, sqHdlcFrame, cbHdlcFrame, &fisu, DEMO_TASK_REF, OTX_TIME_INFINITY); 
    if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
      // Driver IO Request buffer is full. We need to retransmit this frame.
      busyCount++;
      OsaThreadSleep(1);
    } else {
      DemoCheckRc(g_sqHdlcSender[nDev].hHandle, result);

      // Step BSN for every new frame package (MSU+FISU or FISU+FISU) to send
      fisu.bsn++;

      if (coFisusSent == 0) {
        OsaThreadSleep(15);
      } else {
        OsaThreadSleep(rand() % 5);
      }

      if (cbHdlcFrame == 3) {
        coFisusSent++;
      } else {
#if 1
        DemoLogf("%d: Sent MSU# %d (Len=%d)\n", nDev, coMsusSent, cbHdlcFrame);
#endif
        coMsusSent++;
      }

      if ((rand() % 100) < 50) {
        // Prepare an MSU
        cbHdlcFrame = ((coMsusSent % 10) * 10) + 10;

        memcpy(&msu, &fisu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
        msu.li = (OTX_UINT8) (cbHdlcFrame - OTX_HDLC_SS7_CB_FISU);  // Set the Length Indicator
        memcpy(sqHdlcFrame, &msu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
        for (i = OTX_HDLC_SS7_CB_FISU; i < cbHdlcFrame; i++) {
          // Add the payload the HDLC frame buffer
          sqHdlcFrame[i] = (OTX_UINT8) i;
        }
      } else {
        // Prepare a FISU
        memcpy(sqHdlcFrame, &fisu, OTX_HDLC_SS7_CB_FISU);  
        cbHdlcFrame = OTX_HDLC_SS7_CB_FISU;
      }
    }
  }
  stopTime = clock();

  DemoLogf("[%d]Done (BusyCount=%d, time=%2.1f seconds) FISUs=%d MSUs=%d.\n", nDev, busyCount, (double) (stopTime - startTime) / CLOCKS_PER_SEC, coFisusSent, coMsusSent);

  return 0;
}
#endif


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char* formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);
}


/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.

  @comm

*******************************************************************************/
static void DemoCheckRc(
  IN OTX_HANDLE hDevice, 
  IN OTX_RESULT result
)
{
  if (OTX_FAILURE(result)) {
    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32 cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32 eDeviceType = 999;    // Device type for hDevice
      OTX_UINT32 nDeviceNumber = 999;  // Device number for hDevice

      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType);
      }

      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber);
      }

      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }

      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int) nDeviceNumber, msgStr);
#if 0
      {
        // API call to get the result code data
        OTX_UINT32 Result, Severity, SourceId, LineNum;

        OtxResultCode(result, &Result, &Severity, &SourceId, &LineNum);
        DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", Result, SourceId, LineNum, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      }
#endif
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

  @comm

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

  @comm

*******************************************************************************/
void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm* ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
#if defined(__linux__)
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;
#elif defined(WIN32)
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
#endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", ptCurrentDayAndTime->tm_mon + 1, ptCurrentDayAndTime->tm_mday, ptCurrentDayAndTime->tm_hour, ptCurrentDayAndTime->tm_min, ptCurrentDayAndTime->tm_sec, nTimeStampMs.sUint32.nLow);
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
}


/*******************************************************************************

  @func
    Idle Function that check for Events from the OTX Driver

  @comm

*******************************************************************************/
void* DemoIdleFunction(
  void* lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY); 

    do {
#if 0 // Single reads of events
      OtxEventDataS eventData;

      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);
      }
#else // Multiple reads of events
      // since we can have many waiting on us, reduce the number of context switches that will take place

#define DEMO_EVENT_DATA_BUF_SIZE 10
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

 
      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched);
      DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        assert(coEventsFetched <= coEventsFetched);
        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }
      }
#endif
    } while (OTX_RESULT_CODE(result) == OTX_S_OK); 
  }

  return 0;
}


/*******************************************************************************

  @func
    Print DMA statistics.

    These stats are obtained from the driver, and are being used during
    initial release as a debugging feature.  These stats may be eliminated
    in the future.

  @comm

*******************************************************************************/
void* DemoStatsThread(
  void* lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_DATETIME nTimeStamp;

  OTX_UINT32 stat_Starts;          // Number of actual dma transfers initiated
  OTX_UINT32 stat_Completes;       // Number of actual dma transfers completed
  OTX_UINT32 stat_Partial;         // Number of partial transfer due to framemask
  OTX_UINT32 stat_Full;            // Number of count of max dma size transfers
  OTX_UINT32 stat_Aborts;          // Number of actual dma transfers aborts
  OTX_UINT32 stat_Amount;          // Number of amount of data transfered
  OTX_UINT32 stat_Errors;          // Number of count of processing errors
  OTX_UINT32 stat_OUruns;          // Number of amount of over/under runs
  OTX_UINT32 stat_OUrunsUser;      // Number of amount of over/under runs caused by user mode
  OTX_UINT32 stat_DmaBusy;         // Number of count of DMA busy occurances
  OTX_UINT32 stat_Notify;          // Number of logical device notifies
  OTX_UINT32 stat_NotifyPartial;   // Number of logical device notifies - partial

  while (g_bStatsThreadActive) {
    OtxDrvGetCurrentTime(&nTimeStamp);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS, &stat_Starts); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES, &stat_Completes); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL, &stat_Partial); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_FULL, &stat_Full); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS, &stat_Aborts); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT, &stat_Amount); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS, &stat_Errors); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS, &stat_OUruns); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER, &stat_OUrunsUser); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_DMABUSY, &stat_DmaBusy); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY, &stat_Notify); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL, &stat_NotifyPartial); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Recv:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n", stat_Starts, stat_Completes, stat_Partial, stat_Full, stat_Aborts, stat_Amount, stat_Errors, stat_OUruns, stat_OUrunsUser, stat_DmaBusy, stat_Notify, stat_NotifyPartial);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_STARTS, &stat_Starts); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES, &stat_Completes); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL, &stat_Partial); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_FULL, &stat_Full); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS, &stat_Aborts); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT, &stat_Amount); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS, &stat_Errors); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS, &stat_OUruns); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER, &stat_OUrunsUser); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_DMABUSY, &stat_DmaBusy); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY, &stat_Notify); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY_PARTIAL, &stat_NotifyPartial); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Send:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n", stat_Starts, stat_Completes, stat_Partial, stat_Full, stat_Aborts, stat_Amount, stat_Errors, stat_OUruns, stat_OUrunsUser, stat_DmaBusy, stat_Notify, stat_NotifyPartial);

    OsaThreadSleep(1000);
  }

  DemoLogf("DemoStatsThread() terminating.\n");

  return 0;
}


void OnSendBurstData(struct sBurstSender* pSnd, OTX_UINT8* pBuf, OTX_UINT32 cbBuf, OTX_UINT32 nRefId)
{
  OTX_UINT32 nUsed = 0;                                                  
  OTX_UINT32 i;
  OTX_RESULT res;

  for (i = 0; i < cbBuf; i++)
    pBuf[i] = (i % 32) ? 0x7e : 0x7e;

  for (i = 0; i < pSnd->nDevs; i++) {
    nUsed = 0;
    res = OtxHdlcExProcessIO(pSnd->hDev[i], pBuf, cbBuf, &nUsed);
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, res);
  }
}



void OnGetBurstData(
  struct sBurstReceiver *   pRcv,
  OTX_UINT8 *               pBuf,
  OTX_UINT32                cbBuf,
  OTX_UINT64                nFrameRefId,
  OTX_DATETIME              dtFrameRefTime
)
{
  OTX_UINT32 nUsed = 0;                                                  
  OTX_RESULT res;
  static int firsttime = 0;

#ifdef USE_HDLCEX_BULK
  res = OtxHdlcExProcessIOBulkEx(&pRcv->hDev[0], pRcv->nDevs, g_coTimeSlotMask, pBuf, cbBuf, nFrameRefId, dtFrameRefTime, &nUsed);
  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, res);
#else
  {
    OTX_UINT32 i;
    for (i = 0; i < (OTX_UINT32) pRcv->nDevs; i++) {
      nUsed = 0;
      res = OtxHdlcExProcessIOEx(pRcv->hDev[i], pBuf, cbBuf, nFrameRefId, dtFrameRefTime, &nUsed);
      DemoCheckRc(OTX_INVALID_HANDLE_VALUE, res);
    }
  }
#endif
}


/*******************************************************************************

  @func
    Burst Sender Thread Function

  @comm

*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_UINT32 nRefId = 0;
  OTX_RESULT otxRc;
  OTX_UINT32 ui32;
  OTX_UINT8 pData[DEMO_BURST_BUF_SIZE];

  if (nSourceId < 0 || nSourceId> 2) {
    DemoLogf("DemoBurstSender: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstSender[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  memset(pData, 0x7e, sizeof(pData));
  
  OnSendBurstData(&g_hLogicDevs.mBurstSender[nSourceId], pData, sizeof(pData), nRefId);
  nRefId++;
  
  ui32 = sizeof(pData);

  otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, pData, &ui32); 
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);
  
  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
//  assert(ui32 == sizeof(pData));

}
 

/*******************************************************************************

  @func
    Burst Receiver Thread Function

  @comm

*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > (OTX_UINT32) g_demoEnv.coHighways ) {
    DemoLogf("DemoBurstReceiver: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  // Pause receivers for senders to be ready and garbage data in DMA goes off
  //OsaThreadSleep(1000); 

  // first read we want the frame reference id and time
  ui32 =  sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);
  
  
  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
//  assert(ui32 == sizeof(buf));
  
  OnGetBurstData(&g_hLogicDevs.mBurstReceiver[nSourceId], buf, sizeof(buf), nFrameRefId, dtFrameRefTime);

}


/*******************************************************************************

  @func
    Prints a Driver Event.

  @comm

*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS* pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
   case OTX_DEVICE_LI_T1E1:
     DemoPrintTimeStamp(pEventData->m_nTimeStamp);

     // Event from Li0 or Li1
     otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
     if (OTX_SUCCESS(otxRc)) {
       if (pEventData->m_nParam) {
         DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
       } else {
         DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
       }
     } else {
       DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
     }
     break;

   case OTX_DEVICE_BURST:
     {
       switch (pEventData->m_nCode) {
       case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
         DemoBurstReceiver(pEventData->m_nRequestId);
         break;
       case OTX_BURST_SENDER_EC_BUF_DEPLETED:
         DemoBurstSender(pEventData->m_nRequestId);
        break;
       default:
         DemoPrintTimeStamp(pEventData->m_nTimeStamp);
         otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
         if (OTX_SUCCESS(otxRc)) {
           DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
         } else {
           DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
         }
         break;
       }
     }
     break;
     
   case OTX_LDEVICE_HDLC_SENDER:
     otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
     DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
     break;

   default:
     // Unhandled event
     DemoPrintTimeStamp(pEventData->m_nTimeStamp);
     DemoLogf("DemoPrintEvent()  Event from an unknown object 0x%X\n", pEventData->m_hDevice);
     break;
  }
}


/*******************************************************************************

  @func

    This function is a callback function, called from the OtxHdlcEx library 
    when an HDLC frame is available.

  @comm

*******************************************************************************/
void DemoPrintHdlcEvent(OtxEventDataS* pEventData)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];
#ifdef DEMO_HDLC_SENDER
  OTX_RESULT result;
#endif

  switch (pEventData->m_eDeviceType) {
   case OTX_LDEVICE_HDLC_RECEIVER:
     // Event from HLDC Receiver
     if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
       OTX_UINT8 buf[10240];
       OTX_UINT32 nFrameLen;
       OTX_UINT32 nFrameSeq;
       OtxHdlcFrameEndCodeE nFrameEndCode;
       OTX_DATETIME nTimeStamp;
       OtxHdlcSS7FisuS* pFisu;
       OTX_INT32 nDev; 
       int f_showdelta = 0;

       do {

         otxRc = OtxHdlcExReadData(pEventData->m_hDevice, buf, sizeof(buf), &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
         if (OTX_RESULT_CODE(otxRc) != OTX_S_OK) {
            //DemoCheckRc(pEventData->m_hDevice, result);
            break;
         }

         if (pEventData->m_nSourceId < 1024) {
           if (g_coHdlcRxFrames[pEventData->m_nSourceId] % 10000 == 9999) {
             f_showdelta = 1;
           } else {
             f_showdelta = 0;
           }
           g_coHdlcRxFrames[pEventData->m_nSourceId]++;
           g_coHdlcRxBytes[pEventData->m_nSourceId] += nFrameLen;
         } else {
           DemoLogf("HDLC Frame Received with an invalid m_nSourceId (%d)\n", pEventData->m_nSourceId);
         }

         nDev = pEventData->m_nSourceId;
         if (g_bSs7SimulationActive[nDev]) {
           if (nFrameEndCode == 0) {
             OTX_BOOL bErroredFrame = OTX_FALSE;

             pFisu = (OtxHdlcSS7FisuS *) &buf[0];

             if (((OTX_UINT8) pFisu->bsn != g_nPrevBsn[nDev]) && ((OTX_UINT8) pFisu->bsn != ((g_nPrevBsn[nDev] + 1) % 128))) {
               bErroredFrame = OTX_TRUE;
               g_nSs7SimulationErrors[nDev]++;
               DemoLogf("=============v Bsn Error (%d)\n", g_nSs7SimulationErrors[nDev]);
             }

             if (nFrameLen == 3) {
               // FISU
               // No further check
             } else {
               // MSU

               if (nFrameLen != (OTX_UINT32) (((g_nSs7SimulationFramesReceived[nDev] % 10) * 10) + 10)) {
                 // Data length check
                 bErroredFrame = OTX_TRUE;
                 g_nSs7SimulationErrors[nDev]++;
                 DemoLogf("=============v Length Error (%d). Expected %d. Got %d.\n", g_nSs7SimulationErrors[nDev], (((g_nSs7SimulationFramesReceived[nDev] % 10) * 2) + 10), nFrameLen);
               } else {
                 // Payload check
                 for (i = 3; i < nFrameLen && i < 1024; i++) {
                   if (buf[i] != (OTX_UINT8) i) {
                     bErroredFrame = OTX_TRUE;
                     g_nSs7SimulationErrors[nDev]++;
                     DemoLogf("=============v Payload Error (%d). Expected %d. Got %d.\n", g_nSs7SimulationErrors[nDev], i, buf[i]);
                   }
                 }
               }


               g_nSs7SimulationFramesReceived[nDev]++;
             }

             /*if (bErroredFrame) */ {
               //              DemoLogf("%d:Len=%.3d  ", pEventData->m_nSourceId, nFrameLen);
               //              for (i = 0; i < nFrameLen && i < 1024; i++) {
               //                DemoLogf("%.2X ", buf[i]);
               //              }
               //              DemoLogf("\n");
           } {
               OTX_UINT32 nFrameStampLo, nFrameStampHi;
               nFrameStampLo = pEventData->m_nRequestId;
               nFrameStampHi = pEventData->m_nParam;
               DemoPrintTimeStamp(nTimeStamp);
               DemoLogf("*HdlcReceiver[%d] Len=%d Fe=0x%.2X Fc=%08x: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, /*nFrameStampHi,*/ nFrameStampLo);
               for (i = 0; i < nFrameLen && i < 1024; i++) {
                 DemoLogf("%.2X ", buf[i]);
               }
               DemoLogf("\n");
             }
             g_nPrevBsn[nDev] = pFisu->bsn;
           } else {
             g_nSs7SimulationErrors[nDev]++;
             DemoLogf("%d:Len=%.3d  Bad Frame\n", pEventData->m_nSourceId, nFrameLen);
             for (i = 0; i < nFrameLen && i < 1024; i++) {
               DemoLogf("%.2X ", buf[i]);
             }
             DemoLogf("\n");
           }

#ifdef DEMO_HDLC_SENDER

           if (g_nSs7SimulationFramesReceived[nDev] == DEMO_SS7_SIM_FRAMES) {
             OTX_INT32 nLi = g_sqHdlcSender[pEventData->m_nSourceId].nLi;
             DemoPrintTimeStamp(nTimeStamp);
             DemoLogf("%d Received %d frames. Receive SS7 Done. Errors=%d\n", pEventData->m_nSourceId, DEMO_SS7_SIM_FRAMES, g_nSs7SimulationErrors[nDev]);
             g_bSs7SimulationActive[nDev] = OTX_FALSE;
             DemoLogf("%d Stopping FISU Sending for HDLC Sender %d\n", pEventData->m_nSourceId, pEventData->m_nSourceId);
             result = OtxHdlcExStopRepeatData(g_sqHdlcSender[pEventData->m_nSourceId].hHandle); 
             result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], OTX_T1E1_LOOPS_OFF); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
           }
#endif
					} else {
						if (g_bHdlcVerbose) {
							// nTimeStamp                this is the timestamp of the hdlc frame
							// nFrameSeq                 this contains the Frame Reference ID
							// pEventData->m_nRequestId  this contains the base Frame Reference ID for this event
							// pEventData->m_nTimeStamp  this contains the base Frame Reference Timestamp for this event
							DemoLogf("%08x ", nTimeStamp.sUint32.nLow);
							DemoLogf("-HdlcReceiver[%.03d] Len=%.03d Fe=0x%.2X Sq=%.05d Fc=%08x: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq, pEventData->m_nRequestId);
							DemoLogf("%08x ", pEventData->m_nTimeStamp.sUint32.nLow);

#if 1
							for (i = 0; i < nFrameLen && i < 1024; i++) {
								DemoLogf("%.2X ", buf[i]);
							}
#endif

							DemoLogf("\n");
						} else {
							if (f_showdelta) {
								DemoLogf("HdlcReceiver[%d] %d frames\n", pEventData->m_nSourceId, g_coHdlcRxFrames[pEventData->m_nSourceId]);
							}         
							if (nFrameEndCode) {
								g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
								//DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d, Fc=%08x: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq, /* pEventData->m_nRequestId,*/ pEventData->m_nParam);
								//for (i = 0; i < nFrameLen && i < 256; i++) {
								//	DemoLogf("%.2X ", buf[i]);
								//}
								//DemoLogf("\n");
							}
						}
					}

#ifdef USE_HDLCEX_BULK
       } while (0);
#else
       } while (OTX_RESULT_CODE(otxRc) == OTX_S_OK);
#endif

     } else {
       // Unexpected event from this device
       otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
       if (OTX_SUCCESS(otxRc)) {
         DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
       } else {
         DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
       }
     }

     return;


  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("DemoPrintHdlcEvent()  Event from an unknown object 0x%X\n", pEventData->m_hDevice);
  }

}


/*******************************************************************************

  @func

    Start burst devices

  @comm

*******************************************************************************/
void DemoStartBurstDevices(void)
{
  OTX_INT32 i;
  OTX_RESULT result;

  for (i = 0; i < g_demoEnv.coHighways; i++) {
    g_hLogicDevs.mBurstReceiver[i].nSourceId = i;
    g_hLogicDevs.mBurstSender[i].nSourceId = i;

    g_hLogicDevs.mBurstReceiver[i].bActive = OTX_TRUE;
    g_hLogicDevs.mBurstSender[i].bActive = OTX_TRUE;
  }

  // by enabling the Burst device, we start the process of transferring data
  // from the hardware to the driver via DMA.  The data is stored in the Burst
  // device's buffer for later processing to connected logical devices.
  DemoLogf("Enable g_hPhysDevs.hBurst\n");
  result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);

}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;
  static OTX_INT32 nCurrentHdlcDev = 0;
  static OTX_INT32 nCurrentLi = 0;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {
   case '?':
   case 'h':
   case 'H':
     DemoLogf("Cmd  Synopsis\n");
     DemoLogf("===  ========\n");
     DemoLogf(" ?   Display Help text\n");
     DemoLogf(" h   Ditto\n");
     DemoLogf(" q   Quit\n\n");
     DemoLogf("Related to T1/E1  ----------------------\n");
     DemoLogf(" <   Increment current Li Device (currently Li#%d selected\n", nCurrentLi);
     DemoLogf(" >   Decrement current Li Device (currently Li#%d selected\n", nCurrentLi);
     DemoLogf(" l   Query Line Status of all Lis\n");
     DemoLogf(" e   Query Line Status Extended of all Lis\n");
     DemoLogf("Related to HDLC:  ----------------------\n");
#ifdef DEMO_HDLC_SENDER
     DemoLogf(" t   Send an HDLC frame from current HDLC Sender\n");
     DemoLogf(" f   Start sending FISUs (continuosly) from current HDLC Sender\n");
     DemoLogf(" i   Start sending LSSUs (continuosly) from current HDLC Sender\n");
     DemoLogf(" m   Send an MSU from current HDLC Sender\n");
     DemoLogf(" F   Toggle OTX_HDLC_SS7_FILTER_FISU_ALL filter mask on/off for current HDLC Sender\n");
#endif
     DemoLogf(" L   Set HDLC Frame length to be transmitted\n");
     DemoLogf(" +   Set number of consecutive frame to be send from the HDLC Sender\n");
     DemoLogf(" S   Set time to sleep (in ms) between transmitted HDLC Frames (%d)\n", g_nSleepTimeBetweenHdlcFrames);
     DemoLogf(" s   Display the SS7 HDLC receive status of the active HDLC Receiver\n");
     DemoLogf(" n   Increment current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
     DemoLogf(" N   Decrement current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
     DemoLogf(" y   Query HDLC Receiver Fifo diagnostics (for current HDLC device)\n");
     DemoLogf(" p   Reset current HDLC Receiver\n");
     DemoLogf(" v   Toggle verbose logging of received HDLC frames\n");
     DemoLogf(" V   Print counters of received HDLC frames\n");
     DemoLogf(" w   Toggle payload checking of received frames\n");
     DemoLogf(" [   Measure T1/E1 span voltage for current Li\n");
     DemoLogf(" c   Set Clock Source to recover from current Li\n");
     DemoLogf(" C   Set Clock Source to INTERNAL\n");
     break;
 
    // E1/T1 span voltmeter
  case '[':
    {
      OTX_UINT32 nMilliVolts;
            
      DemoLogf("Measure T1/E1 Span Voltage for Li%d: \n", nCurrentLi);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_LINE_VOLTAGE, &nMilliVolts); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("  Voltage = %2.3f V   \n", (double)nMilliVolts/1000);
      
    }
    break;


   case 'w':
      {
       g_bHdlcCheckRxFrames = !g_bHdlcCheckRxFrames;
       DemoLogf("g_bHdlcCheckRxFrames is now %s.\n", g_bHdlcCheckRxFrames ? "TRUE" : "FALSE");
     }
     break;

   case 'v':
      {
       OTX_INT32 i;
       g_bHdlcVerbose = !g_bHdlcVerbose;
       DemoLogf("g_bHdlcVerbose is now %s.\n", g_bHdlcVerbose ? "TRUE" : "FALSE");
       for (i = 0; i < g_demoEnv.coHdlcEx; i++) {
         if (g_coHdlcRxFrames[i]) {
           DemoLogf("g_coHdlcRxFrames[%d] = %d (reset to 0) \n", i, g_coHdlcRxFrames[i]);
           g_coHdlcRxFrames[i] = 0;
           g_coHdlcRxBytes[i] = 0;
           g_coHdlcBadRxFrames[i] = 0;
         }
       }
     }
     break;

   case 'V':
      {
       OTX_INT32 i;

       static clock_t startTime = 0;
       double elapsedTime;
       static OTX_INT32 sqBytesReceived[DEMO_MAX_DEV];

       for (i = 0; i < g_demoEnv.coHdlcEx; i++) {
         if (g_coHdlcRxFrames[i]) {
           if (!startTime) {
             DemoLogf("[%.3d]  Tot bytes=%.10d Tot frames=%.6d (Bad=%.6d)\n", i, g_coHdlcRxBytes[i], g_coHdlcRxFrames[i], g_coHdlcBadRxFrames[i]);
           } else {
             elapsedTime = (double) (clock() - startTime) / CLOCKS_PER_SEC;

             DemoLogf("[%.3d]  Tot bytes=%.10d  %6.1f bytes/sec  Tot frames=%.6d (Bad=%.6d)\n", i, g_coHdlcRxBytes[i], (double) (g_coHdlcRxBytes[i] - sqBytesReceived[i]) / elapsedTime, g_coHdlcRxFrames[i], g_coHdlcBadRxFrames[i]);
           }
         }
       }

       // Need a critical section here
       memcpy(sqBytesReceived, g_coHdlcRxBytes, sizeof(OTX_INT32) * DEMO_MAX_DEV);
       startTime = clock();
     }
     break;

   case 's':
      {
       OTX_UINT32 nFisuCount;   // Number of FISUs received
       OTX_UINT32 nLssuCount;   // Number of LSSUs received
       OTX_UINT32 nMsuCount;    // Number of MSUs received
       OtxHdlcSS7FisuS fisu;     // The last received FISU on the pipe.
       DemoLogf("SS7 Status for HdlcReceiver #%d \n", nCurrentHdlcDev);

       result = OtxHdlcExSS7GetReceiveStatus(g_sqHdlcReceiver[nCurrentHdlcDev].hHandle, &nFisuCount, &nLssuCount, &nMsuCount, &fisu); DemoCheckRc(g_sqHdlcReceiver[nCurrentHdlcDev].hHandle, result);

       if (OTX_SUCCESS(result)) {
         DemoLogf("FISUs received: %lu\n", nFisuCount);
         DemoLogf("LSSUs received: %lu\n", nLssuCount);
         DemoLogf("MSUs  received: %lu\n", nMsuCount);
         DemoLogf("Last FISU received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n", fisu.bsn, fisu.bib, fisu.fsn, fisu.fib, fisu.li);
       }
     }
     break;

   case 'L':
      {
       char strBuf[12];
       DemoLogf("Enter HDLC Frame Length to be transmitted (%d) ", g_cbHdlcFrame0);
       OsaConsoleGetStr(strBuf);
       if (*strBuf) {
         g_cbHdlcFrame0 = (OTX_INT32) strtol(strBuf, NULL, 0);
         DemoLogf("HDLC Frame Length is now set to %d\n", g_cbHdlcFrame0);
       } else {
         DemoLogf("HDLC Frame Length remains unchanged\n");
       }
     }
     break;

   case 'S':
      {
       char strBuf[12];
       DemoLogf("Enter Time to Sleep (in ms) between transmitted HDLC Frames (%d) ", g_nSleepTimeBetweenHdlcFrames);
       OsaConsoleGetStr(strBuf);
       if (*strBuf) {
         g_nSleepTimeBetweenHdlcFrames = (OTX_INT32) strtol(strBuf, NULL, 0);
         DemoLogf("Sleep time is now set to %d ms\n", g_nSleepTimeBetweenHdlcFrames);
       } else {
         DemoLogf("Sleep time remains unchanged\n");
       }
     }
     break;

#ifdef DEMO_HDLC_SENDER
   case 'P':
      {
       OTX_RESULT result;
       OTX_INT32 i;
       OTX_INT32 busyCount = 0;
       clock_t startTime, stopTime;
       OtxHdlcSS7FisuS fisu1;
       OtxHdlcSS7FisuS fisu2;
       OTX_UINT8 sqBuf[10];
       OTX_UINT32 nDev = 0;

       g_BsnCheck = OTX_TRUE;
       g_BsnErrors = 0;

       for (i = 0; i < 10; i++) {
         sqBuf[i] = (OTX_UINT8) i;
       }

       fisu1.bib = 1;
       fisu1.fib = 1;
       fisu1.fsn = 0x7F;
       fisu1.bsn = 0x7F;

       memcpy(&fisu2, &fisu1, sizeof(OtxHdlcSS7FisuS));

       DemoLogf("Sending FISUs from dev%d\n", nDev);

       // Send MSU Frame(s)
       startTime = clock();
       for (i = 0; i < 1000; i++) {
         //result = OtxHdlcExSS7SendData(g_sqHdlcSender[nDev].hHandle, (OTX_UINT8 *)&fisu1, 3, &fisu2, DEMO_TASK_REF, OTX_TIME_INFINITY); 

         result = OtxHdlcExSS7SendData(g_sqHdlcSender[nDev].hHandle, sqBuf, (i == 999) ? 5 : 4, &fisu2, DEMO_TASK_REF, OTX_TIME_INFINITY); 
         if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
           // Driver IO Request buffer is full. We need to retransmit this frame.
           i--;
           busyCount++;
         } else {
           DemoCheckRc(g_sqHdlcSender[nDev].hHandle, result);

           fisu1.bsn++;
           memcpy(&fisu2, &fisu1, sizeof(OtxHdlcSS7FisuS));

           OsaThreadSleep(1);
         }
       }
       stopTime = clock();

       DemoLogf("[%d]Done g_BsnErrors=%d (BusyCount=%d, time=%2.1f seconds).\n", nDev, g_BsnErrors, busyCount, (double) (stopTime - startTime) / CLOCKS_PER_SEC);
     }

     break;
#endif 

   case '+':
      {
       char strBuf[12];

       DemoLogf("Set the number of consecutive frame to sent when 'm', 't', or 'T' are pressed (%d)\n", g_coHdlcSenderFrames);
       OsaConsoleGetStr(strBuf);
       if (*strBuf) {
         g_coHdlcSenderFrames = (OTX_INT32) strtol(strBuf, NULL, 0);
         DemoLogf("Consecutive frame count is now %d\n", g_coHdlcSenderFrames);
       } else {
         DemoLogf("Consecutive frame count remains unchanged\n");
       }
     }
     break;


#ifdef DEMO_HDLC_SENDER
   case 't':
     OsaThreadStartEx(DemoHdlcTransmitFunction, (void *) nCurrentHdlcDev);
     break;

   case 'U':
      {
       // This performs a simulation of MSU and FISU sending. 
       // The Li is looped back to itself. 
       OTX_INT32 nLi = g_sqHdlcSender[nCurrentHdlcDev].nLi;
       g_nSs7SimulationFramesReceived[nCurrentHdlcDev] = 0;
       g_nSs7SimulationErrors[nCurrentHdlcDev] = 0;
       g_nPrevBsn[nCurrentHdlcDev] = 0x7F;


       result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], OTX_T1E1_LOCAL_LOOP); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);

       OsaThreadStartEx(DemoSs7SimulationFunction, (void *) &nCurrentHdlcDev);
     }
     break;


   case 'u':
      {
       // This performs a simulation of MSU and FISU sending. 
       // The Li is looped back to itself. 
       OTX_INT32 nLi = g_sqHdlcSender[nCurrentHdlcDev].nLi;
       static OTX_BOOL g_bLooped[16];

       g_bLooped[nLi] = !g_bLooped[nLi];

       DemoLogf("Li#%d is now %s\n", g_bLooped[nLi] ? "Looped" : "Not looped");

       if (g_bLooped[nLi]) {
         result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], OTX_T1E1_LOCAL_LOOP); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
       } else {
         result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], OTX_T1E1_LOOPS_OFF); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
       }
     }
     break;


   case 'T':
      {
       OTX_UINT8* pBadMsg = (OTX_UINT8*) "\x00\x01\x26\x3a\x4f\x02\x02\x68\x4d\x08\x03\x82\x90\x00";
       OTX_INT32 nMsgLen = 14;
       //OTX_UINT8 *pBadMsg = "\x04\x09\x7F";
       //OTX_INT32 nMsgLen = 3;
       OTX_HANDLE hHandle;

       hHandle = g_sqHdlcSender[nCurrentHdlcDev].hHandle;

       DemoLogf("Sending test HDLC frame from HDLC Sender #%d Li#%d...", nCurrentHdlcDev, g_sqHdlcSender[nCurrentHdlcDev].nLi);

       result = OtxHdlcExSendData(hHandle, pBadMsg, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

       DemoLogf("Done.\n");
     }
     break;

   case 'F':
      {
       OTX_HANDLE hHandle;
       OTX_UINT32 filterMask;

       hHandle = g_sqHdlcReceiver[nCurrentHdlcDev].hHandle;

       result = OtxHdlcExSS7GetFilter(hHandle, &filterMask); DemoCheckRc(hHandle, result);
       if (filterMask & OTX_HDLC_SS7_FILTER_FISU_ALL) {
         filterMask &= ~OTX_HDLC_SS7_FILTER_FISU_ALL;
       } else {
         filterMask |= OTX_HDLC_SS7_FILTER_FISU_ALL;
       }

       DemoLogf("Setting SS7 filter mask to %.4X for HDLC Receiver #%d (Li#%d)...", filterMask, nCurrentHdlcDev, g_sqHdlcReceiver[nCurrentHdlcDev].nLi);

       result = OtxHdlcExSS7SetFilter(hHandle, filterMask); DemoCheckRc(hHandle, result);

       DemoLogf("Done.\n");
     }
     break;

   case 'f':
      {
       OTX_HANDLE hHandle;

       hHandle = g_sqHdlcSender[nCurrentHdlcDev].hHandle;
       DemoLogf("Start Sending FISUs (fsn=%d) continously from HDLC Sender #%d (Li#%d)...", nCurrentHdlcDev, g_sqHdlcSender[nCurrentHdlcDev].nLi);

       if (nCurrentHdlcDev == 0) {
         g_tFisu0.fsn++;
         result = OtxHdlcExStartRepeatData(hHandle, (OTX_UINT8 *) &g_tFisu0, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);
       } else {
         g_tFisu1.fsn++;
         result = OtxHdlcExStartRepeatData(hHandle, (OTX_UINT8 *) &g_tFisu1, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);
       }
       // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures

       DemoLogf("Done.\n");
     }
     break;

   case 'i':
      {
       OTX_HANDLE hHandle;

       hHandle = g_sqHdlcSender[nCurrentHdlcDev].hHandle;

       // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
       g_tLssu.fsn = g_tFisu0.fsn;
       g_tLssu.sf = g_tLssu.sf + 1; 
       g_tLssu.li = 1; // Short LSSU

       DemoLogf("Start Sending LSSUs (fsn=%d, sf=%d) continously from HDLC Sender #%d (Li#%d)...", g_tLssu.fsn, g_tLssu.sf, nCurrentHdlcDev, g_sqHdlcSender[nCurrentHdlcDev].nLi);

       result = OtxHdlcExStartRepeatData(hHandle, (OTX_UINT8 *) &g_tLssu, OTX_HDLC_SS7_CB_SHORT_LSSU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

       DemoLogf("Done.\n");
     }
     break;

   case '.':
      {
       int i;

       for (i = 0; i < g_demoEnv.coHdlcEx; i++) {
         OTX_HANDLE hHandle;

         hHandle = g_sqHdlcSender[i].hHandle;
         DemoLogf("Start Sending FISUs (fsn=%d) continously from HDLC Sender #%d ...\n", i);

         //      if (nCurrentHdlcDev == 0) 
          {
           g_tFisu0.fsn++;
           result = OtxHdlcExStartRepeatData(hHandle, (OTX_UINT8 *) &g_tFisu0, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);
         }
         //      else {
         //        g_tFisu1.fsn++;
         //result = OtxHdlcExStartRepeatData(hHandle, (OTX_UINT8 *) &g_tFisu1, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);
         //      }
         // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
       }

       DemoLogf("Done.\n");
     }
     break;

   case 'p':
      {
       OTX_HANDLE hHandle;

       hHandle = g_sqHdlcReceiver[nCurrentHdlcDev].hHandle;


       DemoLogf("Reset HDLC Receiver #%d (Li#%d)...", nCurrentHdlcDev, g_sqHdlcSender[nCurrentHdlcDev].nLi);

       result = OtxDrvReset(hHandle); DemoCheckRc(hHandle, result);

       DemoLogf("Done.\n");
     }
     break;

   case 'm':
      {
       if (nCurrentHdlcDev == 0) {
         g_tFisu0.fsn++;
       } else {
         g_tFisu1.fsn++;
       }
       OsaThreadStartEx(DemoMsuTransmitFunction, (void *) &nCurrentHdlcDev);
     }
     break;

   case 'r':
      {
       OTX_UINT8* pRepeatFrame = (OTX_UINT8*) "\x10\x11\x12\x13\x14";
       OTX_INT32 nMsgLen = 5;
       OTX_HANDLE hHandle;

       hHandle = g_sqHdlcSender[nCurrentHdlcDev].hHandle;

       DemoLogf("Start Repeat Data Frames from HDLC Sender #%d (Li#%d)...", nCurrentHdlcDev, g_sqHdlcSender[nCurrentHdlcDev].nLi);

       result = OtxHdlcExStartRepeatData(hHandle, pRepeatFrame, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

       DemoLogf("Done.\n");
     }
     break;

   case 'R':
      {
       OTX_HANDLE hHandle;

       hHandle = g_sqHdlcSender[nCurrentHdlcDev].hHandle;

       DemoLogf("Stop Repeat Data Frames from HDLC Sender #%d (Li#%d)...", nCurrentHdlcDev, g_sqHdlcSender[nCurrentHdlcDev].nLi);

       result = OtxHdlcExStopRepeatData(hHandle); DemoCheckRc(hHandle, result);

       DemoLogf("Done.\n");
     }
     break;
#endif


#ifdef DEMO_HDLC_SENDER
   case 'n':
      {
       // Select current nCurrentHdlcDev
       nCurrentHdlcDev++;
       if (nCurrentHdlcDev > (g_demoEnv.coHdlcEx) - 1) {
         // Wrap 
         nCurrentHdlcDev = 0;
       }

       DemoLogf("DSP HDLC Sender#%d (Li#%d, TS%d) Selected \n", nCurrentHdlcDev, g_sqHdlcSender[nCurrentHdlcDev].nLi, g_sqHdlcSender[nCurrentHdlcDev].nTs);
     }
     break;


   case 'N':
      {
       // Select current nCurrentHdlcDev
       nCurrentHdlcDev--;
       if (nCurrentHdlcDev < 0) {
         // Wrap
         nCurrentHdlcDev = (g_demoEnv.coHdlcEx) - 1;
       }

       DemoLogf("DSP HDLC Sender#%d (Li#%d, TS%d) Selected \n", nCurrentHdlcDev, g_sqHdlcSender[nCurrentHdlcDev].nLi, g_sqHdlcSender[nCurrentHdlcDev].nTs);
     }
     break;
#endif

   case '#':
      {
       DemoLogf("Performing an illegal operation. Crash test\n");
       memcpy(NULL, &nCurrentHdlcDev, 200);
     }
     break;


     /*
       case 'R':
         {
           OTX_INT32 nLi;

           for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
             DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
             result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
             DemoLogf("Done.");
           }
         }
         break;
     */
   case ',':
      {
       OTX_INT32 nLi;

       for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
         DemoLogf("Reseting Li #%d... ", nLi);
         result = OtxDrvReset(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
         DemoLogf("Done.");
       }
     }
     break;

  case 'g':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'G':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("CMDR Receiver Reset on Li #%d... ", nLi);
        result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nLi], 2, 0x40); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;




   case 'l':
      {
       OtxT1E1Layer1StatusE eLayer1Status;
       OTX_INT32 nLi;

       for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
         DemoLogf("Line status of Li #%d is ", nLi);
         result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
         DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
       }
     }
     break;

  case 'e':
    {
      OtxT1E1Layer1StatusExS sLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Li #%d line is ", nLi);
        result = OtxT1E1GetLayer1StatusEx(g_hPhysDevs.hLi[nLi], &sLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);

        DemoLogf("Alarms: ");
        if (sLayer1Status.bRedAlarm) {
          DemoLogf("R");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bBlueAlarm) {
          DemoLogf("B");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bYellowAlarm) {
          DemoLogf("Y");
        } else {
          DemoLogf("-");
        }

        if (sLayer1Status.bFrameError) {
          DemoLogf(" FrmErr: YES");
        } else {
          DemoLogf(" FrmErr: NO ");
        }

        if (sLayer1Status.bMultiFrameError) {
          DemoLogf(" MFrmErr: YES");
        } else {
          DemoLogf(" MFrmErr: NO ");
        }
        DemoLogf("\n");
      }
    }
    break;


  case 'd':   // dump falc registers
    DemoDumpFalcRegisters();
    break;


  case 'k':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'K':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in QuadLi%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hQuadLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hQuadLi[nCurrentLi], result);

      DemoLogf("QuadLi%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'o':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'O':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in QuadLi%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hQuadLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hQuadLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to QuadLi%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;


   case '>':
      {
       // Increment current nCurrentLi
       nCurrentLi++;
       if (nCurrentLi > (g_demoEnv.coLi) - 1) {
         // Wrap 
         nCurrentLi = 0;
       }

       DemoLogf(" Li#%d Selected \n", nCurrentLi);
     }
     break;


   case '<':
      {
       // Decrement current nCurrentLi
       nCurrentLi--;
       if (nCurrentLi < 0) {
         // Wrap
         nCurrentLi = g_demoEnv.coLi - 1;
       }

       DemoLogf(" Li#%d Selected \n", nCurrentLi);
     }
     break;


   case '*':
      {
       OTX_UINT32 nValue;
       char sValue[80];

       result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
       DemoLogf("Enter queue notification value (1-32) [currently %d]: ", nValue);
       OsaConsoleGetStr(sValue);
       if (sValue[0] == '\0') {
         break;
       }
       nValue = strtol(sValue, NULL, 0);

       // before setting the new trigger value, the burst device must not be enabled
       result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);
       result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
       result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);

       result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
       DemoLogf("New queue notification value is %d\n", nValue);

       break;
     }

   case '\'':
     { // Read back the threshold
       OTX_UINT32 nRxThreshold;
       result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);
       DemoLogf("Li[%d]: RX Input Threshold = %d\n", nCurrentLi, nRxThreshold);
     }
     break;

  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;

      DemoLogf("Setting Clock Source to recover from current Li (#%d)\n", nCurrentLi);

      eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + nCurrentLi ); 
      result = OtxBrdSetClocks(g_hPhysDevs.hBoardController, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      DemoLogf("Done.\n");
    }
    break;
  
  case 'C':
    {
      OtxBrdClockSourceE  eClkSource;

      DemoLogf("Setting Clock Source to INTERNAL\n");

      eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_INTERNAL ); 
      result = OtxBrdSetClocks(g_hPhysDevs.hBoardController, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(g_hPhysDevs.hBoardController, result);

      DemoLogf("Done.\n");
    }
    break;
  

   case 'q':
   case 'Q':
     return 1;
     break;

   case '~':
      {
       char strBuf[12];
       OTX_INT32 nSleepTime;

       DemoPrintf("Enter milliseconds to sleep :");
       OsaConsoleGetStr(strBuf);

       if (strBuf[0]) {
         nSleepTime = strtol(strBuf, NULL, 0);
         DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
         OsaThreadSleep(nSleepTime);
       }
     }
     break;

   default:
     DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
     break;
  }

  return 0;
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char* argv[], DemoEnvT* envP)
{
  short i;
  char szDeviceNoList[128];

  // Set reasonable defaults before reading from command line
  envP->pszInputStr = NULL;
  envP->bExitOnError = OTX_FALSE;
  envP->eT1E1Mode = OTX_T1E1_MODE_E1;
  envP->boardNo = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coLi = 8;
  envP->coHighways = 2;
  envP->eClkSource        = OTX_CLOCK_SOURCE_LOCAL_0;
  envP->bVerboseLogging = OTX_FALSE;
  envP->bDevEnableList = OTX_FALSE;
  envP->coDevEnableList = 0;

  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->eTermMode = OTX_T1E1_TERM_100_OHM;
  envP->bMonitorMode = OTX_FALSE;
  envP->bAmplifier = OTX_FALSE;
  envP->nRxThreshold = DEMO_DEFAULT_RX_THRESHOLD; // in mV
  envP->bDriveTxPins = OTX_TRUE;

  envP->bSs7FisuFiltering = OTX_TRUE;
  envP->bHdlcBitInversion = OTX_FALSE;
  envP->bSuperChannel = OTX_FALSE;
  envP->b56kbpsChannel = OTX_FALSE;
  envP->nTimeSlots = 1;
  envP->coHdlcPerLi = 31; 
  envP->nStartTS = 1;
  envP->nCapacity = 64;


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {
    // Scan command line arguments
    if (*(argv[i] + 1) == 'Z') {
      // Command input string -- case sensitive
      envP->pszInputStr = argv[i] + 2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {
        // Is argument a legal switch?
        if (*(argv[i] + 1) == 'I') {
          // HDLC Bit Inversion
          envP->bHdlcBitInversion = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'X') {
          // Use Super-channel (higher than 64kbps capacity)
          envP->bSuperChannel = OTX_TRUE;
        } 
        
        /*else if (*(argv[i] + 1) == '5') {
          // Use 56kbps HDLC channel (vs 64kpbs)
          envP->b56kbpsChannel = OTX_TRUE;
		  envP->nCapacity = 56;
        } else if (*(argv[i] + 1) == '3') {
          // Use 32kbps HDLC channel (vs 64kpbs)
		  envP->nCapacity = 32;
        } else if (*(argv[i] + 1) == '6') {
          // Use 16kbps HDLC channel (vs 64kpbs)
		  envP->nCapacity = 16;
        } else if (*(argv[i] + 1) == '8') {
          // Use 16kbps HDLC channel (vs 64kpbs)
		  envP->nCapacity = 8;
        }*/ 

        else if (*(argv[i] + 1) == 'K') {
          // Set capacity of channel
          envP->nCapacity = (strtoul(argv[i] + 2, NULL, 0)) * 8;
          if ((envP->nCapacity < 8) || (envP->nCapacity > 64)) {
            DemoLogf("Error: Invalid capacity of channel (%s bits, %i Kbps)\n", argv[i]+2, envP->nCapacity);          
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i] + 1) == 'W') {
          char *pStr;
          // Enable a specific list of receivers only (by device numbers)
          envP->bDevEnableList = OTX_TRUE;
          envP->coDevEnableList = 0;
          strcpy(szDeviceNoList, argv[i] + 2);

          pStr = strtok(szDeviceNoList, ",:.");

          while (pStr) {
            envP->sqDevEnableList[envP->coDevEnableList] = strtoul(pStr, NULL, 0);
            envP->coDevEnableList++;
            pStr = strtok(NULL, ",:.");
          }

        } else if (*(argv[i] + 1) == 'D') {
          // Number of timeslots in the superchannel
          envP->coHdlcPerLi = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'V') {
          // Verbose Logging
          envP->bVerboseLogging = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'R') {
          // Number of timeslots in the superchannel
          envP->nStartTS = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'N') {
          // Number of timeslots in the superchannel
          envP->nTimeSlots = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'S') {
          // SS7 FISU filtering
          envP->bSs7FisuFiltering = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'L') {
          // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i] + 1) == 'Q') {
          // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'B') {
          // Board number
          envP->boardNo = strtoul(argv[i] + 2, NULL, 0);

        } else if (*(argv[i] + 1) == 'P') {
          // Tri-state transmit pins
          envP->bDriveTxPins = OTX_FALSE;
        } else if (*(argv[i] + 1) == 'A') {
          // Use Front End Amplifier = ON
          envP->bAmplifier = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'E') {
          // Use ReceiveEqualizer  = ON
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i] + 1) == '2') {
          // Use MonitorMode  = ON
          envP->bMonitorMode = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'H') {
          // RxThreshold (in mV)
          envP->nRxThreshold = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'Y') {
          // TermMode
          if ( *(argv[i]+2) == '1') {
            envP->eTermMode = OTX_T1E1_TERM_100_OHM;
          } else if ( *(argv[i]+2) == '7') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM;
          } else if ( *(argv[i]+2) == 'Z') {
            envP->eTermMode = OTX_T1E1_TERM_HIGHZ;
          } else if ( *(argv[i]+2) == 'S') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i] + 1) == 'T') {
          // Board Type
          if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coLi = 4;
            envP->coHighways = 1;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi = 8;
            envP->coHighways = 2;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi = 2;
            envP->coHighways = 2;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi = 1;
            envP->coHighways = 1;
          } else if ( *(argv[i]+2) == 'G') {
            envP->nBoardType = OTX_DEVICE_GIMLE_16_PCI_PLUS;
            envP->coLi = 16;
            envP->coHighways = 4;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'O') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == '2') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else if (*(argv[i]+2) == '6') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_6;
          } else if (*(argv[i]+2) == '7') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_7;
          } else if (*(argv[i]+2) == '8') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_8;
          } else if (*(argv[i]+2) == '9') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_9;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '0') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_10;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '1') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_11;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '2') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_12;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '3') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_13;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '4') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_14;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '5') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_15;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i] + 1) == 'M') {
          // Li Mode
          if (*(argv[i] + 2) == 'E') {
            // E1
            envP->eT1E1Mode = OTX_T1E1_MODE_E1;
          } else if (*(argv[i] + 2) == 'T') {
            // T1
            envP->eT1E1Mode = OTX_T1E1_MODE_T1;
            if (envP->nTimeSlots > 24) {
              envP->nTimeSlots = 24;
            }
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[T|4|P|S|G]: Board Type used for the demo, where\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  S = Use Sleipnir Board\n");
  DemoLogf("                  G = Use Gimle-16-PCI-Plus Board\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n\n");
  DemoLogf(" /S           : SS7 FISU filtering active\n");
  DemoLogf(" /I           : Use HDLC Bit Inversion (Default is to not invert the bits)\n");
  /*
  DemoLogf(" /5           : Use HDLC 56kbps channel.\n");
  DemoLogf(" /3           : Use HDLC 32kbps channel.\n");
  DemoLogf(" /6           : Use HDLC 16kbps channel.\n");
  DemoLogf(" /8           : Use HDLC 8kbps channel.\n");
  */
  DemoLogf(" /K<bits>     : Bit capacity (1-8)\n");
  DemoLogf(" /O[I|x]        : Clock Source, where\n");
  DemoLogf(" /X           : Use HDLC Super-channels (higher than 64kbps capacity). Only for OTXHDLC.\n");
  DemoLogf(" /N<timeslots>: Number of 64kbps timeslots in the Super-channels (only used if /X is used)\n");
  DemoLogf(" /R<timeslots>: Start timeslot in the Super-channels (only used if /X is used)\n");
  DemoLogf(" /A           : Front End Amplifier = ON (default is OFF)\n");
  DemoLogf(" /E           : Receive Equalizer = ON (longhaul) (default is OFF)\n");
  DemoLogf(" /2           : -20dB MonitorMode = ON (default is OFF)\n");
  DemoLogf(" /H           : Receive Input Threshold in mV (default is %d mV)\n", DEMO_DEFAULT_RX_THRESHOLD);
  DemoLogf(" /Y           : Termination Mode, where\n");
  DemoLogf("                  1 = 100/120 ohm balanced\n");
  DemoLogf("                  7 = 75 ohm unbalanced\n");
  DemoLogf("                  Z = High-impedance\n");
  DemoLogf("                  S = 75 ohm balanced\n");
  DemoLogf(" /P           : Tri-state transmit pins (default is drive transmit pins)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
  DemoLogf(" /V           : Verbose Logging\n");
  DemoLogf(" /W<dev>[,:<dev>..]  : Enable specific list of RX devices only\n");
  DemoLogf("                List device numbers separated by \',\'\n");
  DemoLogf(" /F[E|S|9]    : T1 Frame Format, where\n");
  DemoLogf("                  E = ESF (default)\n");
  DemoLogf("                  4 = F4\n");
  DemoLogf("                  S = SF [D3/D4] (F12)\n");
  DemoLogf("                  9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]      : T1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]    : E1 Frame Format, where\n");
  DemoLogf("                  D = Doubleframe\n");
  DemoLogf("                  C = CRC4 Multiframe (default)\n");
  DemoLogf("                  G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]      : E1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  H = HDB3 (default)\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char* formatStr, ...)
{
  va_list argList;

  //EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile, formatStr, argList);
  }

  va_end(argList);

  //LeaveCriticalSection(&g_demoLogfCritSec);
}


/*******************************************************************************

  @func

  prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf2(char* formatStr, ...)
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile, formatStr, argList);
  }

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  OTX_INT32 i;
  for (i = 0; i < g_demoEnv.coHighways; i++) {
    if (g_hLogicDevs.mBurstSender[i].pTxFile) {
      fclose(g_hLogicDevs.mBurstSender[i].pTxFile);
      g_hLogicDevs.mBurstSender[i].pTxFile = NULL;
    }
    if (g_hLogicDevs.mBurstReceiver[i].pRxFile) {
      fclose(g_hLogicDevs.mBurstReceiver[i].pRxFile);
      g_hLogicDevs.mBurstReceiver[i].pRxFile = NULL;
    }
  }

#if 0
  fclose(g_pBurstSendFile);
  fclose(g_pBurstRecvFile);
#endif

  fclose(g_pDemoLogFile);
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char* argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  printf("HdlcExDemo1, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for Burst device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

#if 0
  if ((g_pBurstSendFile = fopen("BurstSend.dat", "wt")) == NULL) {
    OsaConsoleExit();
    exit(-1);
  }

  if ((g_pBurstRecvFile = fopen("BurstRecv.dat", "wt")) == NULL) {
    OsaConsoleExit();
    exit(-1);
  }
#endif  
  atexit(DemoAtExit);

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
   {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static" : "dynamic");
  }


#if 0   // this is NOT needed w/ 1.05 version of OtxHdlcEx library
  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();

  switch (OTX_RESULT_CODE(result)) {
   case OTX_RES_HANDLE:
     DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
     OsaConsoleExit();
     OtxDrvReleaseLib();
     exit(1);
     break;
   case OTX_RES_NOT_SUPPORTED:
     DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
     OsaConsoleExit();
     OtxDrvReleaseLib();
     exit(1);
     break;
   default:
     DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
     break;
  }
#endif

  DemoLogf("Connecting OtxHdlcEx lib...\n");
  result = OtxHdlcExConnectLib();

  switch (OTX_RESULT_CODE(result)) {
   case OTX_RES_HANDLE:
     DemoLogf("Error: Could not connect to the OTX HdlcEx library.\n");
     OsaConsoleExit();
     OtxDrvReleaseLib();
     exit(1);
     break;
   default:
     DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
     break;
  }

  // Init all lobal objects 
  DemoInitializeGlobalData();

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Set clock source for the board
  {
    char szClockSource[64];
    
    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      strcpy(szClockSource, "Internal");
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
    case OTX_CLOCK_SOURCE_LOCAL_1:
    case OTX_CLOCK_SOURCE_LOCAL_2:
    case OTX_CLOCK_SOURCE_LOCAL_3:
    case OTX_CLOCK_SOURCE_LOCAL_4:
    case OTX_CLOCK_SOURCE_LOCAL_5:
    case OTX_CLOCK_SOURCE_LOCAL_6:
    case OTX_CLOCK_SOURCE_LOCAL_7:
    case OTX_CLOCK_SOURCE_LOCAL_8:
    case OTX_CLOCK_SOURCE_LOCAL_9:
    case OTX_CLOCK_SOURCE_LOCAL_10:
    case OTX_CLOCK_SOURCE_LOCAL_11:
    case OTX_CLOCK_SOURCE_LOCAL_12:
    case OTX_CLOCK_SOURCE_LOCAL_13:
    case OTX_CLOCK_SOURCE_LOCAL_14:
    case OTX_CLOCK_SOURCE_LOCAL_15:
      sprintf(szClockSource, "Line Interface %d", (int)g_demoEnv.eClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
      break;
    default:
      strcpy(szClockSource, "Unknown");
      break;
    }
    
    DemoLogf("Clock Source: %s\n", szClockSource);
    result = OtxBrdSetClocks(g_hPhysDevs.hBoardController, OTX_CLOCK_MODE_LOCAL, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hBoardController, result);
  }



  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();

  // Initialized the Connected Logical Devices
  DemoLogf("Initializing devices...\n");
  DemoInitializeLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting logical devices...\n");
  DemoConnectLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  DemoStartBurstDevices();

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // abort threads
  DemoLogf("\nExiting...\n"); {
    OTX_INT32 i;


    g_bStatsThreadActive = OTX_FALSE;
    for (i = 0; i < g_demoEnv.coHighways; i++) {
      g_hLogicDevs.mBurstReceiver[i].bActive = OTX_FALSE;
      g_hLogicDevs.mBurstSender[i].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1500);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();

#if 1   // this is NOT needed w/ 1.05 version of OtxHdlcEx library
  OtxDrvReleaseLib();
#endif

  OtxHdlcExReleaseLib();

  return 0;
}


/*******************************************************************************

  @func

  Dump the contents of the falc registers for the purpose of debugging..

*******************************************************************************/
void DemoDumpFalcRegisters(void)
{
  OTX_INT32 i;
  OTX_UINT32 nAddress, readVal;
  OTX_RESULT result;

  for (i = 0; i < g_demoEnv.coLi; i++) {
    DemoLogf("\nFalc[%d] register dump\n=====================\n", i);
    for (nAddress = 0; nAddress < 0x100; nAddress++) {
      result = OtxDrvReadReg8(g_hPhysDevs.hLi[i], nAddress, &readVal);
      if (OTX_RESULT_CODE(result) != OTX_E_INVALIDARG) {
        DemoCheckRc(g_hPhysDevs.hLi[i], result);
      }
      if ((nAddress & 0xf) == 0x0) {
        DemoLogf("0x%.2X  ", nAddress);
      }
      if ((nAddress & 0xf) == 0x8) {
        DemoLogf("  ");
      }
      DemoLogf(" %.2X", (OTX_UINT8) readVal);
      if ((nAddress & 0xf) == 0xf) {
        DemoLogf("\n");
      }
    }
    if ((nAddress & 0xf) == 0xf) {
      DemoLogf("\n");
    }
  }
}


