/* $Workfile: HdlcExRxManyTs.c $       $Revision: 14 $      $Modtime: 1/03/07 1:34p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    HdlcExRxManyTs.c

    A simple example console application demonstrating the use of the
    software HDLC receivers (no HDLC sending) using the OtxHdlcEx Library.

    The following boards are supported by this application:

			Thor-2(4,8)-PCI-Plus
      Sleipnir-1-PCI-Plus
      Gimle-16-PCI-Plus
      Gimle-16-PCI-Basic-32
      Gimle-16-PCI-Basic-64

    For Gimle-16-PCI-Plus, this demo configures HDLC receivers only on 16 Lines; 

    For Gimle-16-PCI-Basic-32, this demo configures HDLC receivers only on 1 Line (32 timeslots); 
    Note that this is just an arbitrary choice by this demo program, as this board can be configured
    to handle up to 32 timeslots from any timeslot on any of its 16 Lines.

    For Gimle-16-PCI-Basic-64, this demo configures HDLC receivers only on 2 Lines (32 timeslots per line); 
    Note that this is just an arbitrary choice by this demo program, as this board can be configured
    to handle up to 64 timeslots from any timeslot on any of its 16 Lines.

    For Thor-2-PCI-Plus, this demo configures HDLC receivers its 2 Lines (32 timeslots per line); 

    For Thor-4-PCI-Plus, this demo configures HDLC receivers its 4 Lines (32 timeslots per line); 

    For Thor-8-PCI-Plus, this demo configures HDLC receivers its 8 Lines (32 timeslots per line); 

    The direct receives operate on a highway basis and are
    hosted by the Burst device.

    The programs configures 32 64kbps or 56kbps HDLC channels (devices) per Li (TS0 - TS31). 
    However, TS0 is normally not used for HDLC data. It is the synchronization 
    timeslot in framed E1 modes.

    Hdlc Device 0  is connected to Li0:TS0
    Hdlc Device 1  is connected to Li0:TS1
    Hdlc Device 2  is connected to Li0:TS2
    :
    Hdlc Device 32 is connected to Li1:TS0
    Hdlc Device 33 is connected to Li1:TS1
    :
    etc

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

	  Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "../OtxRev.h"
#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxBrd.h"

#include "OtxBurst.h"
#include "OtxBurstD.h"

#include "OtxHdlcD.h"
#include "OtxHdlcEx.h"
#include "OtxData.h"

/*******************************************************************************
 Defines
*******************************************************************************/


#define DEMO_DEFAULT_LOG_FILE      "HdlcExRxManyTs.log"
#define DEMO_DEFAULT_FRAMES_FILE   "HdlcExRxManyTs.frm"
#define DEMO_DEFAULT_BOARD_NO      0
#define DEMO_DEFAULT_RX_THRESHOLD  500     // in mV
#define DEMO_BURST_BUF_SIZE        0x8000    // 32k
#define DEMO_MAX_DEV               512
#define DEMO_MAX_BOARDS            10
#define DEMO_EVENT_DATA_BUF_SIZE   10

/*******************************************************************************
 Structures
*******************************************************************************/
typedef struct DemoEnvT {
  // Common 
  char* pszInputStr;
  OTX_BOOL bExitOnError;
  char cwd[OTX_MAX_PATH];
  char logFileName[OTX_MAX_PATH];
  char framesFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int coLi;
  int coHighways;
  int coHdlcExDevs;
  OtxT1E1LiModeE eT1E1Mode;    // T1 or E1 mode
  int firstBoardNo;           // Board number of the first boards
  int nBoards;           // Number of boards
  OTX_BOOL               bReceiveEqualizer;
  OtxT1E1TermModeE       eTermMode;
  OTX_BOOL               bMonitorMode;
  OTX_BOOL               bAmplifier;
  OTX_INT32              nRxThreshold;
  OTX_BOOL               bDriveTxPins;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OtxBrdClockSourceE     eClkSource;
  OTX_BOOL bVerboseLogging;

  OTX_BOOL bDevEnableList;
  OTX_INT32 sqDevEnableList[128];
  OTX_INT32 coDevEnableList;

  // HdlcEx related
  int coHdlcPerLi;
  OTX_BOOL b56kbpsChannel;
  OTX_BOOL bSs7FisuFiltering;
  OTX_BOOL bSs7FisuFilteringAll;
  OTX_BOOL bHdlcBitInversion;
  OTX_BOOL bFlagsAsInterframeTimeFill;
  OTX_BOOL bSkipHdlcDecodingOnTs0;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(OTX_INT32 nBrd);
static void DemoConfigureE1T1Devices(OTX_INT32 nBrd);
static void DemoEnablePhysicalDevices(OTX_INT32 nBrd);
static void DemoConnectDevices(OTX_INT32 nBrd);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS* pE1CfgData, OUT OtxT1ConfigOptionsS* pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS* pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char* formatStr, ...);
static void DemoLogf2(char* formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char* argv[], DemoEnvT* envP);
static void DemoCreateLogicalDevices(OTX_INT32 nBrd);
static void DemoInitializeLogicalDevices(OTX_INT32 nBrd);
static void DemoEnableLogicalDevices(OTX_INT32 nBrd);
static void DemoPrintHdlcEvent(OtxEventDataS* pEventData);
static void DemoPrintEvent(OtxEventDataS* pEventData);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_UINT32 g_coTimeSlotMask = 32;
OTX_EVENT g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE g_hEventQueue = 0;
FILE* g_pDemoLogFile = NULL;
FILE* g_pDemoFramesFile = NULL;
OsaMutexS g_demoLogfCritSec;

DemoEnvT g_demoEnv;
OTX_BOOL g_bStatsThreadActive = OTX_FALSE;

OTX_DATETIME g_nTimeStampForBytePerSecCalculation;


struct sPhysDevs {
  OTX_HANDLE hBoardController;
  OTX_HANDLE hLi[16];
  OTX_HANDLE hQuadLi[4];
  OTX_HANDLE hBurst;
  int nBoard;
} g_hPhysDevs[DEMO_MAX_BOARDS];

struct sPipes {
  OTX_HANDLE hLiOut[16];
} g_hPipes[DEMO_MAX_BOARDS];

struct sHdlcReceiver {
  HDLCX_HANDLE hHandle;
  OTX_INT32 nLi;
  OTX_UINT32 nTs;
  OTX_UINT32 nTsOnBurstHwy;
  OTX_UINT32 nBurstHwy;

  OTX_UINT32 coHdlcRxBytes;
  OTX_UINT32 coHdlcRxFrames;
  OTX_UINT32 coHdlcBadRxFrames;

};


struct sHdlcRx {
  struct sHdlcReceiver mRx[DEMO_MAX_DEV];  // one for each possible highway hosted by the Burst device
} g_hdlc[DEMO_MAX_BOARDS];


// Burst DMA receiver
struct sBurstReceiver {
  OTX_UINT32 nSourceId;
  OTX_BOOL bActive;
  HDLCX_HANDLE hDev[DEMO_MAX_DEV];
  OTX_UINT32 nDevs;
};

struct sLogicDevs {
  struct sBurstReceiver mBurstReceiver[4];  // one for each possible highway hosted by the Burst device
} g_hLogicDevs[DEMO_MAX_BOARDS];

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************

  @func
    Initalize the global data for the Hdlc Receiver devices, which will later
    be used in OTX API calls.

  @comm

*******************************************************************************/
void DemoInitializeGlobalData(void)
{
  OTX_INT32 i, j, n, nBrd;
  OTX_UINT32 nDivisor;

  if (g_demoEnv.nBoardType != OTX_DEVICE_GIMLE_16_PCI_BASIC_32) {
    g_demoEnv.coHdlcExDevs = g_demoEnv.coLi * g_demoEnv.coHdlcPerLi;
  }

  // Set g_coTimeSlotMask, which is the number of timeslots on one burst highway.
  // This number varies between different board types
  switch (g_demoEnv.nBoardType) {
   case OTX_DEVICE_GIMLE_16_PCI_PLUS:
     g_coTimeSlotMask = 128; 
     break;
   case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
     g_coTimeSlotMask = 32; 
     break;
   case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
     g_coTimeSlotMask = 32; 
     break;
   case OTX_DEVICE_SLEIPNIR_1_PCI:
     g_coTimeSlotMask = 32;
     break;
   case OTX_DEVICE_THOR_8_PCI_PLUS:
     g_coTimeSlotMask = 128;
     break;
   case OTX_DEVICE_THOR_4_PCI_PLUS:
     g_coTimeSlotMask = 128;
     break;
   case OTX_DEVICE_THOR_2_PCI_PLUS:
     g_coTimeSlotMask = 32;
     break;
   default:
     DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
     OsaConsoleExit();
     OtxHdlcExReleaseLib();
     exit(1);
     break;
  }

  if ((OTX_UINT32) g_demoEnv.coHdlcExDevs > g_demoEnv.coHighways * g_coTimeSlotMask) {
    DemoLogf("Error: Too big number of devices. Failed to map timeslots.\n");
    OsaConsoleExit();
    OtxHdlcExReleaseLib();
    exit(1);
  }

  if (g_demoEnv.nBoards > DEMO_MAX_BOARDS) {
    DemoLogf("Error: Too big number of boards.\n");
    OsaConsoleExit();
    OtxHdlcExReleaseLib();
    exit(1);
  }

    
  nDivisor = g_coTimeSlotMask / 32; // Number of E1 spans in one Burst highway

  for (nBrd = 0; nBrd < g_demoEnv.nBoards; nBrd++) {
    n = 0;
    for (i = 0; i < g_demoEnv.coLi; i++) {
      
      for (j = 0; j < g_demoEnv.coHdlcPerLi; j++,n++) {
        
        g_hdlc[nBrd].mRx[n].nLi = i;
        g_hdlc[nBrd].mRx[n].nTs = j;
        g_hdlc[nBrd].mRx[n].nTsOnBurstHwy = j + (i % nDivisor) * 32;
        g_hdlc[nBrd].mRx[n].nBurstHwy = i / nDivisor;
        
        
        if (g_demoEnv.bVerboseLogging) {
          // Print mapping for debug purposes
          DemoLogf("HdlcReceiver[%d:%.3d]: nLi=%d, nTs=%d, nBurstHwy=%d nTsOnBurstHwy=%d\n",nBrd, n, g_hdlc[nBrd].mRx[n].nLi, g_hdlc[nBrd].mRx[n].nTs, g_hdlc[nBrd].mRx[n].nBurstHwy, g_hdlc[nBrd].mRx[n].nTsOnBurstHwy);
        }
      }
    }
  }

}


/*******************************************************************************

  @func

*******************************************************************************/
void DemoPrintSignalSettings(void)
{
  // Signal settings
  char *pszTermMode;

  switch (g_demoEnv.eTermMode) {
  case OTX_T1E1_TERM_120_OHM:
    pszTermMode = "120  ";
    break;
  case OTX_T1E1_TERM_100_OHM:
    pszTermMode = "100  ";
    break;
  case OTX_T1E1_TERM_75_OHM:
    pszTermMode = "75U  ";
    break;
  case OTX_T1E1_TERM_HIGHZ:
    pszTermMode = "HIGHZ";
    break;
  case OTX_T1E1_TERM_75_OHM_BALANCED:
    pszTermMode = "75B  ";
    break;
  default:
    pszTermMode = "???  ";
    break;
  }

  DemoLogf("RxEQ=%s, TermMode=%s, MonMode=%s, Amp=%s, RxThold=%.4d, TxPins=%s\n",
    g_demoEnv.bReceiveEqualizer ? "ON " : "OFF",
    pszTermMode,
    g_demoEnv.bMonitorMode ? "ON " : "OFF",
    g_demoEnv.bAmplifier ? "ON " : "OFF",
    g_demoEnv.nRxThreshold,
    g_demoEnv.bDriveTxPins ? "DRIVE" :"HIGHZ"
  );

}




void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

  @comm

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(2048, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);
}

  
void DemoCreatePhysicalDevices(OTX_INT32 nBrd)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the OTX board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
   case OTX_DEVICE_GIMLE_16_PCI_PLUS:
     DemoLogf("Opening Gimle-16-PCI-Plus board\n");
     break;
   case OTX_DEVICE_GIMLE_16_PCI_BASIC_32:
     DemoLogf("Opening Gimle-16-PCI-Basic-32 board\n");
     break;
   case OTX_DEVICE_GIMLE_16_PCI_BASIC_64:
     DemoLogf("Opening Gimle-16-PCI-Basic-64 board\n");
     break;
   case OTX_DEVICE_SLEIPNIR_1_PCI:
     DemoLogf("Opening Sleipnir board\n");
     break;
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      DemoLogf("Opening Thor-2-PCI Plus board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      DemoLogf("Opening Thor-4-PCI Plus board\n");
      break;
   case OTX_DEVICE_THOR_8_PCI_PLUS:
     DemoLogf("Opening Thor-8-PCI Plus board\n");
     break;
   default:
     DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
     OsaConsoleExit();
     exit(1);
     break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.firstBoardNo + nBrd, g_demoEnv.firstBoardNo + nBrd, g_hEventQueue, &g_hPhysDevs[nBrd].hBoardController); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the 2 individual T1/E1 line interface members within the Quad device
      // Note that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[nBrd].hBoardController, OTX_DEVICE_LI_T1E1, i, i | (nBrd << 8), g_hEventQueue, &g_hPhysDevs[nBrd].hLi[i]); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
    for (i = 0; i < 1; i++) {
      // Get handles to the one quad T1/E1 line interface devices
      // Note that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[nBrd].hBoardController, OTX_DEVICE_QUAD_T1E1, i, i | (nBrd << 8), g_hEventQueue, &g_hPhysDevs[nBrd].hQuadLi[i]); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      // Note that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[nBrd].hQuadLi[i], OTX_DEVICE_LI_T1E1, j, (j + i*4) | (nBrd << 8), g_hEventQueue, &g_hPhysDevs[nBrd].hLi[j + i*4]); DemoCheckRc(g_hPhysDevs[nBrd].hQuadLi[i], result);
      }
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the two quad T1/E1 line interface devices
      // Note that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[nBrd].hBoardController, OTX_DEVICE_QUAD_T1E1, i, i | (nBrd << 8), g_hEventQueue, &g_hPhysDevs[nBrd].hQuadLi[i]); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      // Note that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[nBrd].hQuadLi[i], OTX_DEVICE_LI_T1E1, j, (j + i * 4) | (nBrd << 8), g_hEventQueue, &g_hPhysDevs[nBrd].hLi[j + i * 4]); DemoCheckRc(g_hPhysDevs[nBrd].hQuadLi[i], result);
      }
    }
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI) {
    // Open the Line interfaces
      // Note that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[nBrd].hBoardController, OTX_DEVICE_LI_T1E1, i, i | (nBrd << 8), g_hEventQueue, &g_hPhysDevs[nBrd].hLi[i]); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);
    }

    // we'll source the Falc from the Burst device, and source the burst device from the FALC
    //  Falc <-- Burst Send (P2S)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[nBrd].hBoardController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);
    //  Burst Recv (S2P) <-- Falc
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[nBrd].hBoardController, OTX_ATTR_BRD_S2P_SOURCED, OTX_BRD_S2P_SOURCED_FALC); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);

	} else if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_BASIC_32 || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_BASIC_64 ) {
    for (i = 0; i < 4; i++) {
      // Get handles to the four quad T1/E1 line interface devices
      // Note that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[nBrd].hBoardController, OTX_DEVICE_QUAD_T1E1, i, i | (nBrd << 8), g_hEventQueue, &g_hPhysDevs[nBrd].hQuadLi[i]); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      // Note that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[nBrd].hQuadLi[i], OTX_DEVICE_LI_T1E1, j, (j + i * 4) | (nBrd << 8), g_hEventQueue, &g_hPhysDevs[nBrd].hLi[j + i * 4]); DemoCheckRc(g_hPhysDevs[nBrd].hQuadLi[i], result);
      }
    }
	}

  // Get a handle to the physical Burst device.
      // Note that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs[nBrd].hBoardController, OTX_DEVICE_BURST, 0,  0 | (nBrd << 8), g_hEventQueue, &g_hPhysDevs[nBrd].hBurst); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);

  // Burst TX interrupts will be turned off when the Burst device is enabled (for boards that support this action)
  #if ( OTX_DRV_REV_NO_MINOR >= 4 )  // driver version 2.4.0.1 and up supports this attribute  
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_SEND_DISABLED,  OTX_TRUE); 
  #endif 

}

  
/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
static void DemoConfigureE1T1Devices(OTX_INT32 nBrd)
{
  OTX_RESULT result;
  OTX_INT32 i;

  DemoPrintSignalSettings();

  // Set default configuration data
  DemoInitT1E1ConfigData(&(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));


  for (i = 0; i < g_demoEnv.coLi; i++) {
     
    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs[nBrd].hLi[i], g_demoEnv.eTermMode); DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);
		    
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1
      // Configure T1 Transceivers
      result = OtxT1SetConfiguration(g_hPhysDevs[nBrd].hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);
      
      // Set them in T1 mode
      result = OtxT1E1SetLiMode(g_hPhysDevs[nBrd].hLi[i], OTX_T1E1_MODE_T1); DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);
    } else {
      // E1
      // Configure E1 Transceivers
      result = OtxE1SetConfiguration(g_hPhysDevs[nBrd].hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);
      
      // Set them in E1 mode
      result = OtxT1E1SetLiMode(g_hPhysDevs[nBrd].hLi[i], OTX_T1E1_MODE_E1); DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);
    }
    
    // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
    DemoLogf("Li[%d:%d]: Setting XmitPinControl to %s\n", nBrd, i, g_demoEnv.bDriveTxPins ? "DRIVE" : "TRISTATE");
    result = OtxT1E1XmitPinControl(g_hPhysDevs[nBrd].hLi[i], g_demoEnv.bDriveTxPins); DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);

    // Set receive threshold
    DemoLogf("Li[%d:%d]: Setting RX Input Threshold to %d\n", nBrd, i, g_demoEnv.nRxThreshold);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[nBrd].hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nRxThreshold); DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);
    { // Read back the threshold
      OTX_UINT32 nRxThreshold;
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);
      DemoLogf("Li[%d:%d]: RX Input Threshold = %d\n", nBrd, i, nRxThreshold);
    }

    // Set Monitor (-20dB) mode
    DemoLogf("Li[%d:%d]: Setting Monitor mode to %s\n", nBrd, i, g_demoEnv.bMonitorMode ? "ON" : "OFF");
    result = OtxT1E1SetMonitorMode(g_hPhysDevs[nBrd].hLi[i], g_demoEnv.bMonitorMode); DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);

    // Turning front end Amplifier ON/OFF      
    DemoLogf("Li[%d:%d]: Setting Front End Amplifier to %s\n", nBrd, i, g_demoEnv.bAmplifier ? "ON" : "OFF");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs[nBrd].hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bAmplifier); DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);
    
  }



}


void DemoEnablePhysicalDevices(OTX_INT32 nBrd)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

  @comm

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs[%d].hBoardController\n", nBrd);
  result = OtxDrvEnable(g_hPhysDevs[nBrd].hBoardController, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
    for (i = 0; i < 1; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs[%d].hQuadLi[%d]\n", nBrd, i);
      result = OtxDrvEnable(g_hPhysDevs[nBrd].hQuadLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[nBrd].hQuadLi[i], result);
    }
  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs[%d].hQuadLi[%d]\n", nBrd, i);
      result = OtxDrvEnable(g_hPhysDevs[nBrd].hQuadLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[nBrd].hQuadLi[i], result);
    }
  }

	if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_BASIC_32 || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_BASIC_64 ) {
    for (i = 0; i < 4; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs[%d].hQuadLi[%d]\n", nBrd, i);
      result = OtxDrvEnable(g_hPhysDevs[nBrd].hQuadLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[nBrd].hQuadLi[i], result);
    }
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Enable Physical Devices, Li's
    DemoLogf("Enable g_hPhysDevs[%d].hLi[%d]\n", nBrd, i);
    result = OtxDrvEnable(g_hPhysDevs[nBrd].hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[nBrd].hLi[i], result);
  }

}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(OTX_INT32 nBrd)
{
  OTX_INT32 nLi;
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];
  OTX_INT32 nTs;

  DemoLogf("%d: Creating pipe for Burst Receiver\n", nBrd);

  if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_BASIC_32) {
    // Create the Pipes between the Burst Device and Li0 (this is a demo)
    for (nLi = 0; nLi < 1; nLi++) {

    // we must define the highway to use 
    iHighwayNo = 0;

    // Create pipes (one for each direction) Use all timeslots except TS0.
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask) * sizeof(sqBurstTimeSlotMask[0]));

    // 32 timeslots
    for (nTs = 0; nTs < 32; nTs++) {
      sqTimeSlotMask[nTs] = 0xFF;
      sqBurstTimeSlotMask[nTs] = 0xFF;
    }

    // --------------------------------------------------------------------
    // Create an incoming and outgoing Pipe representing the external T1/E1
    // --------------------------------------------------------------------
    result = OtxDrvCreatePipe(64 * 32, &g_hPipes[nBrd].hLiOut[nLi]);
    DemoCheckRc(g_hPipes[nBrd].hLiOut[nLi], result);

    // -------------------------
    // Connect pipe: Li -> Burst
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes[nBrd].hLiOut[nLi], OTX_PIPE_INPUT, g_hPhysDevs[nBrd].hLi[nLi], -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes[nBrd].hLiOut[nLi], result);

    result = OtxDrvConnectPipe(g_hPipes[nBrd].hLiOut[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs[nBrd].hBurst, iHighwayNo, sqBurstTimeSlotMask, g_coTimeSlotMask);
    DemoCheckRc(g_hPipes[nBrd].hLiOut[nLi], result);
  }
	
} else { 

  // Create the Pipes (one for each direction) between the Burst Device and the LIs
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {

    // we must define the highway to use 
    iHighwayNo = nLi / (g_demoEnv.coLi / g_demoEnv.coHighways);

    // Create pipes (one for each direction) Use all timeslots except TS0.
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask) * sizeof(sqBurstTimeSlotMask[0]));

    for (nTs = 0; nTs < 32; nTs++) {
      sqTimeSlotMask[nTs] = 0xFF;

      // for this demo, we'll lay out the burst time slot mask as follows:
      //
      // OTX_DEVICE_THOR_8_PCI_PLUS:
      // for this demo, we'll lay out the burst time slot mask as follows:
      //    [ 0..31]  - LI0:TS0-TS31 and LI4:TS0-TS31
      //    [32..63]  - LI1:TS0-TS31 and LI5:TS0-TS31
      //    [64..95]  - LI2:TS0-TS31 and LI6:TS0-TS31
      //    [96..127] - LI3:TS0-TS31 and LI7:TS0-TS31
      //
      // upon connecting the pipe (see below), we have to pass the highway
      // number (0-3 - one highway per each QuadFalc (128TS per highway)
      // as well as the size of the highway (128 TS)
      //
      // for this demo, we'll lay out the burst time slot mask as follows:
      //    [ 0..31]  - LI0:TS0-TS31 and LI4:TS0-TS31
      // OTX_DEVICE_THOR_8_PCI_PLUS, OTX_DEVICE_GIMLE_16_PCI_PLUS only:
      //    [32..63]  - LI1:TS0-TS31 and LI5:TS0-TS31
      //    [64..95]  - LI2:TS0-TS31 and LI6:TS0-TS31
      //    [96..127] - LI3:TS0-TS31 and LI7:TS0-TS31
      //
      // upon connecting the pipe (see below), we have to pass the highway
      // number (0-3 - one highway per each QuadFalc (128TS per highway)
      // as well as the size of the highway (128 TS)
      //
      // OTX_DEVICE_SLEIPNIR_1_PCI:
      //    [ 0..31]  - LI0:TS0-TS31
      //
      // OTX_DEVICE_HUGIN_2_PC104:
      // OTX_DEVICE_THOR_2_PCI_PLUS:
      //    [ 0..31]  - LI0:TS0-TS31 (hwy0) and LI1:TS0-TS31 (hwy1) 
      // 
      // OTX_DEVICE_THOR_8_PCI_PLUS:
      //    [ 0..31]  - LI0:TS0-TS31 (hwy0)  and LI4:TS0-TS31 (hwy1) 
      //    [32..63]  - LI1:TS0-TS31 (hwy0)  and LI5:TS0-TS31 (hwy1) 
      //    [64..95]  - LI2:TS0-TS31 (hwy0)  and LI6:TS0-TS31 (hwy1) 
      //    [96..127] - LI3:TS0-TS31 (hwy0)  and LI7:TS0-TS31 (hwy1) 
      //
      // OTX_DEVICE_GIMLE_16_PCI_PLUS:
      //    [ 0..31]  - LI0:TS0-TS31 (hwy0)  and LI4:TS0-TS31 (hwy1)  and LI8:TS0-TS31  (hwy2)  and LI8:TS0-TS31  (hwy3)
      //    [32..63]  - LI1:TS0-TS31 (hwy0)  and LI5:TS0-TS31 (hwy1)  and LI9:TS0-TS31  (hwy2)  and LI9:TS0-TS31  (hwy3)
      //    [64..95]  - LI2:TS0-TS31 (hwy0)  and LI6:TS0-TS31 (hwy1)  and LI10:TS0-TS31 (hwy2)  and LI10:TS0-TS31 (hwy3)
      //    [96..127] - LI3:TS0-TS31 (hwy0)  and LI7:TS0-TS31 (hwy1)  and LI11:TS0-TS31 (hwy2)  and LI11:TS0-TS31 (hwy3)
      //
      // For OTX_DEVICE_THOR_8_PCI_PLUS and OTX_DEVICE_GIMLE_16_PCI_PLUS 
      // upon connecting the pipe (see below), we have to pass the highway
      // number (0,1 for Thor-8 and 0,1,2,3 for Gimle-16). There is one highway 
      // per each QuadFalc (128TS per highway). We also pass the size of the highway,
      // which is 128 (TS) for these two boards.
      //
      // For OTX_DEVICE_SLEIPNIR_1_PCI, OTX_DEVICE_HUGIN_2_PC104, and 
      // OTX_DEVICE_THOR_2_PCI_PLUS, the size of the highway is 32.
      //
      // this is just a simple algorithm to accomplish the above
      sqBurstTimeSlotMask[nTs + ((nLi % (g_demoEnv.coLi / g_demoEnv.coHighways)) * 32)] = 0xFF;
    }

      // --------------------------------------------------------------------
      // Create an incoming and outgoing Pipe representing the external T1/E1
      // --------------------------------------------------------------------
      result = OtxDrvCreatePipe(64 * 32, &g_hPipes[nBrd].hLiOut[nLi]);
      DemoCheckRc(g_hPipes[nBrd].hLiOut[nLi], result);

      // -------------------------
      // Connect pipe: Li -> Burst
      // -------------------------
      result = OtxDrvConnectPipe(g_hPipes[nBrd].hLiOut[nLi], OTX_PIPE_INPUT, g_hPhysDevs[nBrd].hLi[nLi], -1, sqTimeSlotMask, 32);
      DemoCheckRc(g_hPipes[nBrd].hLiOut[nLi], result);

      result = OtxDrvConnectPipe(g_hPipes[nBrd].hLiOut[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs[nBrd].hBurst, iHighwayNo, sqBurstTimeSlotMask, g_coTimeSlotMask);
      DemoCheckRc(g_hPipes[nBrd].hLiOut[nLi], result);

    }
  }
}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

  @comm

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS* pE1CfgData, 
  OUT OtxT1ConfigOptionsS* pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits = 0x3;
  pE1CfgData->bfSaBits = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe = OTX_TRUE;
  pE1CfgData->bAutoResync = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion = OTX_TRUE;
  pE1CfgData->bAutoRemoteAlarm = OTX_FALSE;
  pE1CfgData->nLosSensitivity = 10;
  pE1CfgData->nLosRecovery = 21;
  pE1CfgData->nLineLength = 10;
  pE1CfgData->bHighTxPower = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion = OTX_FALSE;
  pT1CfgData->eLfaSensitivity = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm = OTX_FALSE;
  pT1CfgData->nLosSensitivity = 10;
  pT1CfgData->nLosRecovery = 21;
  pT1CfgData->nLineLength = 10;
  pT1CfgData->bHighTxPower = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels = 0x00;
}


/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(OTX_INT32 nBrd)
{
  OTX_RESULT result;
  OTX_INT32 i, n;

  for (i = 0; i < g_demoEnv.coHdlcExDevs; i++) {

    // Note that the HDLC Receiver is not using an event queue. Instead we register a call back function for the events.
    // Note also that we OR in the board number in the source ID. In that way we can extract the board number when the receive up events from this device
    result = OtxHdlcExCreate(OTX_LDEVICE_HDLC_RECEIVER, OTX_INVALID_HANDLE_VALUE, i | (nBrd << 16), &g_hdlc[nBrd].mRx[i].hHandle);  DemoCheckRc(g_hdlc[nBrd].mRx[i].hHandle, result);

    // Register the callback function for the HDLC receiver events
    result = OtxHdlcExSetCallback(g_hdlc[nBrd].mRx[i].hHandle, &DemoPrintHdlcEvent); DemoCheckRc(g_hdlc[nBrd].mRx[i].hHandle, result);

    // Fill burst device mapping table
    n = g_hdlc[nBrd].mRx[i].nBurstHwy;  
    g_hLogicDevs[nBrd].mBurstReceiver[n].hDev[g_hLogicDevs[nBrd].mBurstReceiver[n].nDevs++] = g_hdlc[nBrd].mRx[i].hHandle;
  }
}



/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectLogicalDevices(OTX_INT32 nBrd)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[128];

  if (g_demoEnv.b56kbpsChannel) {
    DemoLogf("%d: Configuring 56kbps HDLC channels\n", nBrd);
  } else {
    DemoLogf("%d: Configuring 64kbps HDLC channels\n", nBrd);
  }

  for (i = 0; i < g_demoEnv.coHdlcExDevs; i++) {

    memset(sqTimeSlotMask, 0x00, g_coTimeSlotMask);
    if (g_demoEnv.b56kbpsChannel) {
      sqTimeSlotMask[g_hdlc[nBrd].mRx[i].nTsOnBurstHwy] = 0xFE; // 0xFE is the mask for a 56kpbs capacity pipe
    } else {
      sqTimeSlotMask[g_hdlc[nBrd].mRx[i].nTsOnBurstHwy] = 0xFF; // 0xFF is the mask for a 64kpbs capacity pipe
    }

    result = OtxHdlcExSetTimeSlotMask(g_hdlc[nBrd].mRx[i].hHandle, g_coTimeSlotMask, sqTimeSlotMask);      
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
  }
}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

  @comm

*******************************************************************************/
void DemoInitializeLogicalDevices(OTX_INT32 nBrd)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coHdlcExDevs; i++) {
    // Initialize the Hdlc Receiver 
    result = OtxHdlcExReceiverInit(
      g_hdlc[nBrd].mRx[i].hHandle, // Handle to logical HDLC Filter Device
      g_demoEnv.bFlagsAsInterframeTimeFill ? OTX_HDLC_FFT_FLAGS : OTX_HDLC_FFT_ALL_ONES, // Fill character(s) to be send between HDLC frames
      1,                              // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,                 // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      g_demoEnv.bHdlcBitInversion,    // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                       // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
      OTX_HDLC_TR_NA,                 // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_TRUE                        // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
    ); 
    DemoCheckRc(g_hdlc[nBrd].mRx[i].hHandle, result);          

    if (g_demoEnv.bSs7FisuFiltering) {
      if (g_demoEnv.bSs7FisuFilteringAll) {
        result = OtxHdlcExSS7SetFilter(g_hdlc[nBrd].mRx[i].hHandle, OTX_HDLC_SS7_FILTER_FISU_ALL | OTX_HDLC_SS7_FILTER_SHORT_LSSU_ALL | OTX_HDLC_SS7_FILTER_LONG_LSSU_ALL); 
        DemoCheckRc(g_hdlc[nBrd].mRx[i].hHandle, result);
      } else {
        result = OtxHdlcExSS7SetFilter(g_hdlc[nBrd].mRx[i].hHandle, OTX_HDLC_SS7_FILTER_FISU | OTX_HDLC_SS7_FILTER_SHORT_LSSU | OTX_HDLC_SS7_FILTER_LONG_LSSU); 
        DemoCheckRc(g_hdlc[nBrd].mRx[i].hHandle, result);
      } 
    }
  }
}
  

/*******************************************************************************

  @func
    Enable Configured Logical Devices

  @comm

*******************************************************************************/
void DemoEnableLogicalDevices(OTX_INT32 nBrd)
{
  OTX_RESULT result;
  OTX_INT32 i;

  if (g_demoEnv.bDevEnableList) {
    // Enable specific receivers only as specified in the sqDevEnableList
    for (i = 0; i < g_demoEnv.coDevEnableList; i++) {
      DemoLogf("Enabling HDLC receiver [%d:%d]\n", nBrd, g_demoEnv.sqDevEnableList[i]);
      result = OtxHdlcExEnable(g_hdlc[nBrd].mRx[g_demoEnv.sqDevEnableList[i]].hHandle); DemoCheckRc(g_hdlc[nBrd].mRx[g_demoEnv.sqDevEnableList[i]].hHandle, result);
    }
  } else {
    // Enable all receivers
    DemoLogf("%d: Enabling %d HDLC receivers\n", nBrd, g_demoEnv.coHdlcExDevs);
    for (i = 0; i < g_demoEnv.coHdlcExDevs; i++) {

      if ((g_hdlc[nBrd].mRx[i].nTs == 0) && g_demoEnv.bSkipHdlcDecodingOnTs0) {
        // Unless specifically set by the bSkipHdlcDecodingOnTs0 variable, don't decode TS0
        continue;
      }

      result = OtxHdlcExEnable(g_hdlc[nBrd].mRx[i].hHandle); DemoCheckRc(g_hdlc[nBrd].mRx[i].hHandle, result);
    }
  }
}
  

/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char* formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);
}


/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.

  @comm

*******************************************************************************/
static void DemoCheckRc(
  IN OTX_HANDLE hDevice, 
  IN OTX_RESULT result
)
{
  if (OTX_FAILURE(result)) {
    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32 cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32 eDeviceType = 999;    // Device type for hDevice
      OTX_UINT32 nDeviceNumber = 999;  // Device number for hDevice

      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType);
      }

      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber);
      }

      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }

      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int) nDeviceNumber, msgStr);
#if 0
      {
        // API call to get the result code data
        OTX_UINT32 Result, Severity, SourceId, LineNum;

        OtxResultCode(result, &Result, &Severity, &SourceId, &LineNum);
        DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", Result, SourceId, LineNum, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      }
#endif
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

  @comm

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

  @comm

*******************************************************************************/
void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampUs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm* ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
#if defined(__linux__)
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;
#elif defined(WIN32)
  nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
#endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nTimeStampUs.nUint64 = nTimeStamp.nUint64 / 10 - (nTimeStampSec.nUint64 * 1000000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1, 
      ptCurrentDayAndTime->tm_mday, 
      ptCurrentDayAndTime->tm_hour, 
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec, 
      nTimeStampMs.sUint32.nLow);
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
}


/*******************************************************************************

  @func
    Idle Function that check for Events from the OTX Driver

  @comm

*******************************************************************************/
void* DemoIdleFunction(
  void* lpParameter   // thread data
)
{
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY); 

    do {
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;

 
      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched);
      DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        assert(coEventsFetched <= coEventsFetched);
        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK); 
  }

  return 0;
}


/*******************************************************************************

  @func
    Print DMA statistics.

    These stats are obtained from the driver, and are being used during
    initial release as a debugging feature.  These stats may be eliminated
    in the future.

  @comm

*******************************************************************************/
void* DemoStatsThread(
  void* lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_DATETIME nTimeStamp;
  OTX_INT32 nBrd;

  OTX_UINT32 stat_Starts;          // Number of actual dma transfers initiated
  OTX_UINT32 stat_Completes;       // Number of actual dma transfers completed
  OTX_UINT32 stat_Partial;         // Number of partial transfer due to framemask
  OTX_UINT32 stat_Full;            // Number of count of max dma size transfers
  OTX_UINT32 stat_Aborts;          // Number of actual dma transfers aborts
  OTX_UINT32 stat_Amount;          // Number of amount of data transfered
  OTX_UINT32 stat_Errors;          // Number of count of processing errors
  OTX_UINT32 stat_OUruns;          // Number of amount of over/under runs
  OTX_UINT32 stat_OUrunsUser;      // Number of amount of over/under runs caused by user mode
  OTX_UINT32 stat_DmaBusy;         // Number of count of DMA busy occurances
  OTX_UINT32 stat_Notify;          // Number of logical device notifies
  OTX_UINT32 stat_NotifyPartial;   // Number of logical device notifies - partial

  while (g_bStatsThreadActive) {
    OtxDrvGetCurrentTime(&nTimeStamp);

    for (nBrd = 0; nBrd < g_demoEnv.nBoards; nBrd++) {
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS, &stat_Starts); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES, &stat_Completes); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL, &stat_Partial); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_FULL, &stat_Full); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS, &stat_Aborts); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT, &stat_Amount); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS, &stat_Errors); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS, &stat_OUruns); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER, &stat_OUrunsUser); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_DMABUSY, &stat_DmaBusy); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY, &stat_Notify); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL, &stat_NotifyPartial); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      
      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf("%d: Burst Recv:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n", nBrd, stat_Starts, stat_Completes, stat_Partial, stat_Full, stat_Aborts, stat_Amount, stat_Errors, stat_OUruns, stat_OUrunsUser, stat_DmaBusy, stat_Notify, stat_NotifyPartial);
      
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_STARTS, &stat_Starts); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES, &stat_Completes); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL, &stat_Partial); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_FULL, &stat_Full); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS, &stat_Aborts); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT, &stat_Amount); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS, &stat_Errors); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS, &stat_OUruns); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER, &stat_OUrunsUser); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_DMABUSY, &stat_DmaBusy); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY, &stat_Notify); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nBrd].hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY_PARTIAL, &stat_NotifyPartial); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);
      
      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf("%d: Burst Send:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d\n", nBrd, stat_Starts, stat_Completes, stat_Partial, stat_Full, stat_Aborts, stat_Amount, stat_Errors, stat_OUruns, stat_OUrunsUser, stat_DmaBusy, stat_Notify, stat_NotifyPartial);
    }

    OsaThreadSleep(1000);
  }

  DemoLogf("DemoStatsThread() terminating.\n");

  return 0;
}


/*******************************************************************************

  @func
    Burst Sender Thread Function

  @comm

*******************************************************************************/
void DemoBurstSender(OTX_INT32 nBrd, OTX_UINT32 nSourceId)
{
  OTX_UINT32 nRefId = 0;
  OTX_RESULT otxRc;
  OTX_UINT32 ui32;
  OTX_UINT8 pData[DEMO_BURST_BUF_SIZE];

  // Send dummy data
  memset(pData, 0x7e, sizeof(pData));
  
  ui32 = sizeof(pData);

  otxRc = OtxBurstSendData(g_hPhysDevs[nBrd].hBurst, nSourceId, pData, &ui32); 
  DemoCheckRc(g_hPhysDevs[nBrd].hBurst, otxRc);

}
 

/*******************************************************************************

  @func

*******************************************************************************/
void OnGetBurstData(
  struct sBurstReceiver *   pRcv,
  OTX_UINT8 *               pBuf,
  OTX_UINT32                cbBuf,
  OTX_UINT64                nFrameRefId,
  OTX_DATETIME              dtFrameRefTime
)
{
  OTX_UINT32 nUsed = 0;                                                  
  OTX_RESULT res;
  static int firsttime = 0;

  {
    OTX_UINT32 i;
    for (i = 0; i < (OTX_UINT32) pRcv->nDevs; i++) {
      nUsed = 0;
      res = OtxHdlcExProcessIOEx(pRcv->hDev[i], pBuf, cbBuf, nFrameRefId, dtFrameRefTime, &nUsed);
      DemoCheckRc(OTX_INVALID_HANDLE_VALUE, res);
    }
  }
}


/*******************************************************************************

  @func
    Burst Receiver Function. When burst data is available, this function
    passes the data on to the OtxHdlcEx library for HDLC decoding.

  @comm

*******************************************************************************/
void DemoBurstReceiver(OTX_INT32 nBrd, OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > (OTX_UINT32) g_demoEnv.coHighways ) {
    DemoLogf("%d:DemoBurstReceiver: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nBrd, nSourceId);
    return;
  }

  if (!g_hLogicDevs[nBrd].mBurstReceiver[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  // first read we want the frame reference id and time
  ui32 =  sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs[nBrd].hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs[nBrd].hBurst, otxRc);
  
  
  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  //  assert(ui32 == sizeof(buf));
  
  OnGetBurstData(&g_hLogicDevs[nBrd].mBurstReceiver[nSourceId], buf, sizeof(buf), nFrameRefId, dtFrameRefTime);

}


/*******************************************************************************

  @func
    Prints a Driver Event.

  @comm

*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS* pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
   case OTX_DEVICE_LI_T1E1:
     DemoPrintTimeStamp(pEventData->m_nTimeStamp);

     // Event from Li0 or Li1
     otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
     if (OTX_SUCCESS(otxRc)) {
       if (pEventData->m_nParam) {
         DemoLogf("E1/T1[%d:%d] Event: %s (%d)\n", pEventData->m_nSourceId >> 8, pEventData->m_nSourceId & 0xFF, msgStr, pEventData->m_nParam);
       } else {
         DemoLogf("E1/T1[%d:%d] Event: %s\n", pEventData->m_nSourceId >> 8, pEventData->m_nSourceId & 0xFF, msgStr);
       }
     } else {
       DemoLogf("E1/T1[%d:%d] Invalid Event: %ld \n", pEventData->m_nSourceId >> 8, pEventData->m_nSourceId & 0xFF, pEventData->m_nCode);
     }
     break;

   case OTX_DEVICE_BURST:
     {
       switch (pEventData->m_nCode) {
       case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
         DemoBurstReceiver(pEventData->m_nSourceId >> 8, pEventData->m_nRequestId);
         break;
       case OTX_BURST_SENDER_EC_BUF_DEPLETED:
         DemoBurstSender(pEventData->m_nSourceId >> 8, pEventData->m_nRequestId);
        break;
       default:
         DemoPrintTimeStamp(pEventData->m_nTimeStamp);
         otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
         if (OTX_SUCCESS(otxRc)) {
           DemoLogf("Burst[%d:%d] Event: %s (%d)\n", pEventData->m_nSourceId >> 8, pEventData->m_nSourceId & 0xFF, msgStr, pEventData->m_nParam);
         } else {
           DemoLogf("Burst[%d:%d] Invalid Event: %ld \n", pEventData->m_nSourceId >> 8, pEventData->m_nSourceId & 0xFF, pEventData->m_nCode);
         }
         break;
       }
     }
     break;
     
   default:
     // Unhandled event
     DemoPrintTimeStamp(pEventData->m_nTimeStamp);
     DemoLogf("DemoPrintEvent()  Event from an unknown object 0x%X\n", pEventData->m_hDevice);
     break;
  }
}


/*******************************************************************************

  @func

    This function is a callback function, called from the OtxHdlcEx library 
    when an HDLC frame is available.

  @comm

*******************************************************************************/
void DemoPrintHdlcEvent(OtxEventDataS* pEventData)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];
       OTX_INT32 nDev; 
       OTX_INT32 nBrd;

  switch (pEventData->m_eDeviceType) {
   case OTX_LDEVICE_HDLC_RECEIVER:
     // Event from HLDC Receiver

     // We have encoded the board number in the upper 16 bits of the m_nSourceId
     nDev = pEventData->m_nSourceId & 0xFFFF;
     nBrd = pEventData->m_nSourceId >> 16;

     if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
       OTX_UINT8 buf[10240];
       OTX_UINT32 nFrameLen;
       OTX_UINT32 nFrameSeq;
       OtxHdlcFrameEndCodeE nFrameEndCode;
       OTX_DATETIME nTimeStamp;
       int f_showdelta = 0;

       do {

         otxRc = OtxHdlcExReadData(pEventData->m_hDevice, buf, sizeof(buf), &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
         if (OTX_RESULT_CODE(otxRc) != OTX_S_OK) {
            //DemoCheckRc(pEventData->m_hDevice, result);
            break;
         }

         if (nBrd < 0 || nBrd > g_demoEnv.nBoards) {
           DemoLogf("HDLC Frame Received with an invalid Board Number (%d) [pEventData->m_nSourceId=0x%.8X]\n", nDev, pEventData->m_nSourceId);
           continue;
         }

         if (nDev < 1024) {
           if (g_hdlc[nBrd].mRx[nDev].coHdlcRxFrames % 10000 == 9999) {
             f_showdelta = 1;
           } else {
             f_showdelta = 0;
           }
           g_hdlc[nBrd].mRx[nDev].coHdlcRxFrames++;
           g_hdlc[nBrd].mRx[nDev].coHdlcRxBytes += nFrameLen;
         } else {
           DemoLogf("HDLC Frame Received with an invalid Device Number (%d) [pEventData->m_nSourceId=0x%.8X]\n", nDev, pEventData->m_nSourceId);
           continue;
         }


         if (g_pDemoFramesFile) {
           fprintf(g_pDemoFramesFile, "%08X %08X [%d:%.03d] Len=%.03d Fe=0x%.2X Sq=%.05d ", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow, nBrd, nDev, nFrameLen, nFrameEndCode, nFrameSeq);
           for (i = 0; i < nFrameLen && i < 1024; i++) {
             fprintf(g_pDemoFramesFile, "%.2X ", buf[i]);
           }
           fprintf(g_pDemoFramesFile, "\n");
         }
         
         if (g_demoEnv.bVerboseLogging) {
           // nTimeStamp                this is the timestamp of the hdlc frame
           // nFrameSeq                 this contains the Frame Reference ID
           // pEventData->m_nRequestId  this contains the base Frame Reference ID for this event
           // pEventData->m_nTimeStamp  this contains the base Frame Reference Timestamp for this event
#if 0
           DemoLogf("%08x ", nTimeStamp.sUint32.nLow);
#else
           DemoPrintTimeStamp(nTimeStamp);
#endif
           DemoLogf("-HdlcReceiver[%d:%.03d] Len=%.03d Fe=0x%.2X Sq=%.05d Fc=%08x: ", nBrd, nDev, nFrameLen, nFrameEndCode, nFrameSeq, pEventData->m_nRequestId);
           
#if 1
           for (i = 0; i < nFrameLen && i < 1024; i++) {
             DemoLogf("%.2X ", buf[i]);
           }
#endif
           
           DemoLogf("\n");
         } else {
           if (f_showdelta) {
             DemoLogf("HdlcReceiver[%d:%d] %d frames\n", nBrd, nDev, g_hdlc[nBrd].mRx[nDev].coHdlcRxFrames);
           }         
           if (nFrameEndCode) {
             g_hdlc[nBrd].mRx[nDev].coHdlcBadRxFrames++;
             //DemoLogf("HdlcReceiver[%d:%d] Len=%d, Fe=0x%.2X, Sq=%d, Fc=%08x: ", nBrd, nDev, nFrameLen, nFrameEndCode, nFrameSeq, /* pEventData->m_nRequestId,*/ pEventData->m_nParam);
             //for (i = 0; i < nFrameLen && i < 256; i++) {
             //	DemoLogf("%.2X ", buf[i]);
             //}
             //DemoLogf("\n");
           }
         }
					

       } while (OTX_RESULT_CODE(otxRc) == OTX_S_OK);

     } else {
       // Unexpected event from this device
       otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
       if (OTX_SUCCESS(otxRc)) {
         DemoLogf("HdlcReceiver[%d:%d] Event: %s \n", nBrd, nDev, msgStr);
       } else {
         DemoLogf("HdlcReceiver[%d:%d] Invalid Event: %ld \n", nBrd, nDev, pEventData->m_nCode);
       }
     }

     return;


  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("DemoPrintHdlcEvent()  Event from an unknown object 0x%X\n", pEventData->m_hDevice);
  }

}


/*******************************************************************************

  @func

    Start burst devices

  @comm

*******************************************************************************/
void DemoStartBurstDevices(OTX_INT32 nBrd)
{
  OTX_INT32 i;
  OTX_RESULT result;

  for (i = 0; i < g_demoEnv.coHighways; i++) {
    g_hLogicDevs[nBrd].mBurstReceiver[i].nSourceId = i;
    g_hLogicDevs[nBrd].mBurstReceiver[i].bActive = OTX_TRUE;
  }

  // by enabling the Burst device, we start the process of transferring data
  // from the hardware to the driver via DMA.  The data is stored in the Burst
  // device's buffer for later processing to connected logical devices.
  DemoLogf("Enable g_hPhysDevs[nBrd].hBurst\n");
  result = OtxDrvEnable(g_hPhysDevs[nBrd].hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[nBrd].hBurst, result);

}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  OTX_RESULT result;
  static OTX_INT32 nCurrentHdlcDev = 0;
  static OTX_INT32 nCurrentLi = 0;
  static OTX_INT32 nCurrentBrd = 0;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {
   case '?':
   case 'h':
   case 'H':
     DemoLogf("Cmd  Synopsis\n");
     DemoLogf("===  ========\n");
     DemoLogf(" ?   Display Help text\n");
     DemoLogf(" h   Ditto\n");
     DemoLogf(" q   Quit\n\n");
     DemoLogf("Related to T1/E1  ----------------------\n");
     DemoLogf(" <   Increment current Li Device (currently Li#%d selected\n", nCurrentLi);
     DemoLogf(" >   Decrement current Li Device (currently Li#%d selected\n", nCurrentLi);
     DemoLogf(" m   Increment current Board (currently Board#%d selected\n", nCurrentBrd);
     DemoLogf(" M   Decrement current Board (currently Board#%d selected\n", nCurrentBrd);
     DemoLogf(" l   Query Line Status of all Lis\n");
     DemoLogf(" e   Query Line Status Extended of all Lis\n");
     DemoLogf("Related to HDLC:  ----------------------\n");
     DemoLogf(" s   Display the SS7 HDLC receive status of the active HDLC Receiver\n");
     DemoLogf(" n   Increment current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
     DemoLogf(" N   Decrement current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
     DemoLogf(" y   Query HDLC Receiver Fifo diagnostics (for current HDLC device)\n");
     DemoLogf(" p   Reset current HDLC Receiver\n");
     DemoLogf(" v   Toggle verbose logging of received HDLC frames\n");
     DemoLogf(" V   Print counters of received HDLC frames\n");
     DemoLogf(" [   Measure T1/E1 span voltage for current Li\n");
     DemoLogf(" c   Set Clock Source to recover from current Li\n");
     DemoLogf(" C   Set Clock Source to INTERNAL\n");
     DemoLogf(" ]   Prompt driver to synch driver clock with System Clock\n");
     break;

     // E1/T1 span voltmeter
  case '[':
    {
      OTX_UINT32 nMilliVolts;
            
      DemoLogf("Measure T1/E1 Span Voltage for Li%d: \n", nCurrentLi);
      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], OTX_ATTR_T1E1_LINE_VOLTAGE, &nMilliVolts); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], result);

      DemoLogf("  Voltage = %2.3f V   \n", (double)nMilliVolts/1000);
      
    }
    break;

  
  case 'v':
      {
       OTX_INT32 i;
       g_demoEnv.bVerboseLogging = !g_demoEnv.bVerboseLogging;
       DemoLogf("Logging is now %s.\n", g_demoEnv.bVerboseLogging ? "VERBOSE" : "SILENT");
       for (i = 0; i < g_demoEnv.coHdlcExDevs; i++) {
         if (g_hdlc[nCurrentBrd].mRx[i].coHdlcRxFrames) {
           DemoLogf("coHdlcRxFrames[%d] = %d (reset to 0) \n", i, g_hdlc[nCurrentBrd].mRx[i].coHdlcRxFrames);
           g_hdlc[nCurrentBrd].mRx[i].coHdlcRxFrames = 0;
           g_hdlc[nCurrentBrd].mRx[i].coHdlcRxBytes = 0;
           g_hdlc[nCurrentBrd].mRx[i].coHdlcBadRxFrames = 0;
         }
       }
     }
     break;

   case 'V':
      {
       OTX_INT32 i;

       static OTX_INT32 sqBytesReceived[DEMO_MAX_BOARDS][DEMO_MAX_DEV];

       for (i = 0; i < g_demoEnv.coHdlcExDevs; i++) {
         if (g_hdlc[nCurrentBrd].mRx[i].coHdlcRxFrames) {
           OTX_DATETIME elapsedTime;
           OTX_UINT32 bitsPerSecond;
           OTX_UINT32 elapsedBits;
           OTX_DATETIME elapsedSeconds;
           OTX_DATETIME nCurrentTime;

           OtxDrvGetCurrentTime(&nCurrentTime);
           
           elapsedTime.nUint64 = nCurrentTime.nUint64 - g_nTimeStampForBytePerSecCalculation.nUint64;
           elapsedBits = (g_hdlc[nCurrentBrd].mRx[i].coHdlcRxBytes - sqBytesReceived[nCurrentBrd][i]) * 8;
           elapsedSeconds.nUint64 = (elapsedTime.nUint64 / 10000000);
           if (elapsedSeconds.sUint32.nLow == 0) {
             bitsPerSecond = 0;
           } else {
             bitsPerSecond = elapsedBits / elapsedSeconds.sUint32.nLow;
           }
           
           DemoLogf("[%.3d] Frames=%.6d (Bad=%.6d) Bytes=%.10d [%d bits/sec]\n", i, g_hdlc[nCurrentBrd].mRx[i].coHdlcRxFrames, g_hdlc[nCurrentBrd].mRx[i].coHdlcBadRxFrames, g_hdlc[nCurrentBrd].mRx[i].coHdlcRxBytes, bitsPerSecond);
         }
       }

       // Need a critical section here
       for (i = 0; i < DEMO_MAX_DEV; i++) {
         sqBytesReceived[nCurrentBrd][i] = g_hdlc[nCurrentBrd].mRx[i].coHdlcRxBytes;
       }
       OtxDrvGetCurrentTime(&g_nTimeStampForBytePerSecCalculation);
     }
     break;

   case 's':
      {
       OTX_UINT32 nFisuCount;   // Number of FISUs received
       OTX_UINT32 nLssuCount;   // Number of LSSUs received
       OTX_UINT32 nMsuCount;    // Number of MSUs received
       OtxHdlcSS7FisuS fisu;     // The last received FISU on the pipe.
       DemoLogf("SS7 Status for HdlcReceiver #%d \n", nCurrentHdlcDev);

       result = OtxHdlcExSS7GetReceiveStatus(g_hdlc[nCurrentBrd].mRx[nCurrentHdlcDev].hHandle, &nFisuCount, &nLssuCount, &nMsuCount, &fisu); DemoCheckRc(g_hdlc[nCurrentBrd].mRx[nCurrentHdlcDev].hHandle, result);

       if (OTX_SUCCESS(result)) {
         DemoLogf("FISUs received: %lu\n", nFisuCount);
         DemoLogf("LSSUs received: %lu\n", nLssuCount);
         DemoLogf("MSUs  received: %lu\n", nMsuCount);
         DemoLogf("Last FISU received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n", fisu.bsn, fisu.bib, fisu.fsn, fisu.fib, fisu.li);
       }
     }
     break;

   case 'p':
      {
       OTX_HANDLE hHandle;

       hHandle = g_hdlc[nCurrentBrd].mRx[nCurrentHdlcDev].hHandle;


       DemoLogf("Reset HDLC Receiver #%d (Li#%d)...", nCurrentHdlcDev, g_hdlc[nCurrentBrd].mRx[nCurrentHdlcDev].nLi);

       result = OtxDrvReset(hHandle); DemoCheckRc(hHandle, result);

       DemoLogf("Done.\n");
     }
     break;

   case 'n':
      {
       // Select current nCurrentHdlcDev
       nCurrentHdlcDev++;
       if (nCurrentHdlcDev > (g_demoEnv.coHdlcExDevs) - 1) {
         // Wrap 
         nCurrentHdlcDev = 0;
       }

       DemoLogf("DSP HDLC Receiver#%d (Li#%d, TS%d) Selected \n", nCurrentHdlcDev, g_hdlc[nCurrentBrd].mRx[nCurrentHdlcDev].nLi, g_hdlc[nCurrentBrd].mRx[nCurrentHdlcDev].nTsOnBurstHwy);
     }
     break;


   case 'N':
      {
       // Select current nCurrentHdlcDev
       nCurrentHdlcDev--;
       if (nCurrentHdlcDev < 0) {
         // Wrap
         nCurrentHdlcDev = (g_demoEnv.coHdlcExDevs) - 1;
       }

       DemoLogf("DSP HDLC Receiver#%d (Li#%d, TS%d) Selected \n", nCurrentHdlcDev, g_hdlc[nCurrentBrd].mRx[nCurrentHdlcDev].nLi, g_hdlc[nCurrentBrd].mRx[nCurrentHdlcDev].nTsOnBurstHwy);
     }
     break;

   case 'm':
      {
       // Select current nCurrentBrd
       nCurrentBrd++;
       if (nCurrentBrd > (g_demoEnv.nBoards) - 1) {
         // Wrap 
         nCurrentBrd = 0;
       }

       DemoLogf("Board#%d Selected \n", nCurrentBrd);
     }
     break;


   case 'M':
      {
       // Select current nCurrentBrd
       nCurrentBrd--;
       if (nCurrentBrd < 0) {
         // Wrap
         nCurrentBrd = (g_demoEnv.nBoards) - 1;
       }

       DemoLogf("Board#%d Selected\n", nCurrentBrd);
     }
     break;

  case 'g':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs[nCurrentBrd].hLi[nLi]); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'G':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("CMDR Receiver Reset on Li #%d... ", nLi);
        result = OtxDrvWriteReg8(g_hPhysDevs[nCurrentBrd].hLi[nLi], 2, 0x40); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

   case 'l':
      {
       OtxT1E1Layer1StatusE eLayer1Status;
       OTX_INT32 nLi;

       for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
         DemoLogf("Line status of Li #%d is ", nLi);
         result = OtxT1E1GetLayer1Status(g_hPhysDevs[nCurrentBrd].hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nLi], result);
         DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
       }
     }
     break;

  case 'e':
    {
      OtxT1E1Layer1StatusExS sLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Li #%d line is ", nLi);
        result = OtxT1E1GetLayer1StatusEx(g_hPhysDevs[nCurrentBrd].hLi[nLi], &sLayer1Status); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nLi], result);

        DemoLogf("Alarms: ");
        if (sLayer1Status.bRedAlarm) {
          DemoLogf("R");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bBlueAlarm) {
          DemoLogf("B");
        } else {
          DemoLogf("-");
        }
        if (sLayer1Status.bYellowAlarm) {
          DemoLogf("Y");
        } else {
          DemoLogf("-");
        }

        if (sLayer1Status.bFrameError) {
          DemoLogf(" FrmErr: YES");
        } else {
          DemoLogf(" FrmErr: NO ");
        }

        if (sLayer1Status.bMultiFrameError) {
          DemoLogf(" MFrmErr: YES");
        } else {
          DemoLogf(" MFrmErr: NO ");
        }
        DemoLogf("\n");
      }
    }
    break;

  case 'k':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'K':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in QuadLi%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs[nCurrentBrd].hQuadLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hQuadLi[nCurrentLi], result);

      DemoLogf("QuadLi%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'o':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'O':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in QuadLi%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs[nCurrentBrd].hQuadLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hQuadLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to QuadLi%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;


   case '>':
      {
       // Increment current nCurrentLi
       nCurrentLi++;
       if (nCurrentLi > (g_demoEnv.coLi) - 1) {
         // Wrap 
         nCurrentLi = 0;
       }

       DemoLogf(" Li#%d Selected \n", nCurrentLi);
     }
     break;


   case '<':
      {
       // Decrement current nCurrentLi
       nCurrentLi--;
       if (nCurrentLi < 0) {
         // Wrap
         nCurrentLi = g_demoEnv.coLi - 1;
       }

       DemoLogf(" Li#%d Selected \n", nCurrentLi);
     }
     break;


   case '*':
      {
       OTX_UINT32 nValue;
       char sValue[80];

       result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nCurrentBrd].hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hBurst, result);
       DemoLogf("Enter queue notification value (1-32) [currently %d]: ", nValue);
       OsaConsoleGetStr(sValue);
       if (sValue[0] == '\0') {
         break;
       }
       nValue = strtol(sValue, NULL, 0);

       // before setting the new trigger value, the burst device must not be enabled
       result = OtxDrvDisable(g_hPhysDevs[nCurrentBrd].hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hBurst, result);
       result = OtxDrvSetAttributeValueUint32(g_hPhysDevs[nCurrentBrd].hBurst, OTX_ATTR_BURST_TRIGGER, nValue); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hBurst, result);
       result = OtxDrvEnable(g_hPhysDevs[nCurrentBrd].hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hBurst, result);

       result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nCurrentBrd].hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hBurst, result);
       DemoLogf("New queue notification value is %d\n", nValue);

       break;
     }

   case '\'':
     { // Read back the threshold
       OTX_UINT32 nRxThreshold;
       result = OtxDrvGetAttributeValueUint32(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hLi[nCurrentLi], result);
       DemoLogf("Li[%d]: RX Input Threshold = %d\n", nCurrentLi, nRxThreshold);
     }
     break;


  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;

      DemoLogf("Setting Clock Source to recover from current Li (#%d)\n", nCurrentLi);

      eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_LOCAL_0 + nCurrentLi ); 
      result = OtxBrdSetClocks(g_hPhysDevs[nCurrentBrd].hBoardController, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hBoardController, result);

      DemoLogf("Done.\n");
    }
    break;
  
  case 'C':
    {
      OtxBrdClockSourceE  eClkSource;

      DemoLogf("Setting Clock Source to INTERNAL\n");

      eClkSource = (OtxBrdClockSourceE) ( OTX_CLOCK_SOURCE_INTERNAL ); 
      result = OtxBrdSetClocks(g_hPhysDevs[nCurrentBrd].hBoardController, OTX_CLOCK_MODE_LOCAL, eClkSource); DemoCheckRc(g_hPhysDevs[nCurrentBrd].hBoardController, result);

      DemoLogf("Done.\n");
    }
    break;
  

  case ']':
    {
      OTX_RESULT result;  // OTX Standard Return Code		
      OTX_BOOL bTimeStampFrom1601;
      
      result = OtxDrvGetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, &bTimeStampFrom1601);
      DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
      OtxDrvSetAttributeValueBool(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_TIMESTAMP_FROM_1601, bTimeStampFrom1601);
      DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    }
    break;

   case 'q':
   case 'Q':
     return 1;
     break;

   case '~':
      {
       char strBuf[12];
       OTX_INT32 nSleepTime;

       DemoPrintf("Enter milliseconds to sleep :");
       OsaConsoleGetStr(strBuf);

       if (strBuf[0]) {
         nSleepTime = strtol(strBuf, NULL, 0);
         DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
         OsaThreadSleep(nSleepTime);
       }
     }
     break;

   default:
     DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
     break;
  }

  return 0;
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char* argv[], DemoEnvT* envP)
{
  short i;
  char szDeviceNoList[1024];

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->firstBoardNo      = DEMO_DEFAULT_BOARD_NO;
  envP->nBoards           = 1;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coLi              = 8;
  envP->coHighways        = 2;
  envP->eClkSource        = OTX_CLOCK_SOURCE_LOCAL_0;
  envP->bVerboseLogging   = OTX_TRUE;
  envP->bDevEnableList    = OTX_FALSE;
  envP->coDevEnableList   = 0;

  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->eTermMode         = OTX_T1E1_TERM_100_OHM;
  envP->bMonitorMode      = OTX_FALSE;
  envP->bAmplifier        = OTX_FALSE;
  envP->nRxThreshold      = DEMO_DEFAULT_RX_THRESHOLD; // in mV
  envP->bDriveTxPins      = OTX_TRUE;

  envP->b56kbpsChannel             = OTX_FALSE;
  envP->bFlagsAsInterframeTimeFill = OTX_TRUE;
  envP->bSs7FisuFiltering          = OTX_TRUE;
  envP->bSs7FisuFilteringAll       = OTX_FALSE;
  envP->bHdlcBitInversion          = OTX_FALSE;
  envP->coHdlcPerLi                = 32; 
  envP->bSkipHdlcDecodingOnTs0     = OTX_TRUE;


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);

  envP->framesFileName[0] = '\0';

  for (i = 1; i < argc; i++) {
    // Scan command line arguments
    if (*(argv[i] + 1) == 'Z') {
      // Command input string -- case sensitive
      envP->pszInputStr = argv[i] + 2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {
        // Is argument a legal switch?
        if (*(argv[i] + 1) == 'I') {
          // HDLC Bit Inversion
          envP->bHdlcBitInversion = OTX_TRUE;
        } else if (*(argv[i]+1) == '5') {   // Use 56kbps HDLC channel (vs 64kpbs)
          envP->b56kbpsChannel = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'W') {
          // Enable a specific list of receivers only (by device numbers)
          char *pStr;
          envP->bDevEnableList = OTX_TRUE;
          envP->coDevEnableList = 0;
          strcpy(szDeviceNoList, argv[i] + 2);

          pStr = strtok(szDeviceNoList, ",:.");

          while (pStr) {
            envP->sqDevEnableList[envP->coDevEnableList] = strtoul(pStr, NULL, 0);
            envP->coDevEnableList++;
            pStr = strtok(NULL, ",:.");
          }

        } else if (*(argv[i] + 1) == 'V') {
          // Non-Verbose Logging (can also be changed interactively with the 'v' key
          envP->bVerboseLogging = OTX_FALSE;
        } else if (*(argv[i] + 1) == 'N') {
          // Decode HDLC on TS0 as well
          envP->bSkipHdlcDecodingOnTs0 = OTX_FALSE;
        } else if (*(argv[i] + 1) == 'S') {
          // Type of FISU filtering
          if ( *(argv[i]+2) == 'N') {
            // NO filtering at all
            envP->bSs7FisuFiltering = OTX_FALSE;
            envP->bSs7FisuFilteringAll = OTX_FALSE;
          } else if ( *(argv[i]+2) == 'S') {
            envP->bSs7FisuFiltering = OTX_TRUE;
            envP->bSs7FisuFilteringAll = OTX_FALSE;
          } else if ( *(argv[i]+2) == 'A') {
            envP->bSs7FisuFiltering = OTX_TRUE;
            envP->bSs7FisuFilteringAll = OTX_TRUE;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i] + 1) == 'K') {
          // HDLC Interframe Timefill Character (Flags or AllOnes)
          if ( *(argv[i]+2) == 'F') {
            envP->bFlagsAsInterframeTimeFill = OTX_TRUE;
          } else if ( *(argv[i]+2) == 'A') {
            envP->bFlagsAsInterframeTimeFill = OTX_FALSE;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i] + 1) == 'L') {
          // Info Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i] + 1) == 'R') {
          // Frames Log file
          strcpy(envP->framesFileName, argv[i] + 2);
        } else if (*(argv[i] + 1) == 'Q') {
          // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'B') {
          // Board number
          envP->firstBoardNo = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'C') {
          // Number of boards
          envP->nBoards = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'P') {
          // Tri-state transmit pins
          envP->bDriveTxPins = OTX_FALSE;
        } else if (*(argv[i] + 1) == 'A') {
          // Use Front End Amplifier = ON
          envP->bAmplifier = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'E') {
          // Use ReceiveEqualizer  = ON
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i] + 1) == '2') {
          // Use MonitorMode  = ON
          envP->bMonitorMode = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'H') {
          // RxThreshold (in mV)
          envP->nRxThreshold = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'Y') {
          // TermMode
          if ( *(argv[i]+2) == '1') {
            envP->eTermMode = OTX_T1E1_TERM_100_OHM;
          } else if ( *(argv[i]+2) == '7') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM;
          } else if ( *(argv[i]+2) == 'Z') {
            envP->eTermMode = OTX_T1E1_TERM_HIGHZ;
          } else if ( *(argv[i]+2) == 'S') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i] + 1) == 'T') {
          // Board Type
          if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coLi = 4;
            envP->coHighways = 1;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi = 8;
            envP->coHighways = 2;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi = 2;
            envP->coHighways = 2;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi = 1;
            envP->coHighways = 1;
          } else if ( *(argv[i]+2) == 'G') {
            envP->nBoardType = OTX_DEVICE_GIMLE_16_PCI_PLUS;
            envP->coLi = 16;
            envP->coHighways = 4;
          } else if ( *(argv[i]+2) == '3') {
            envP->nBoardType = OTX_DEVICE_GIMLE_16_PCI_BASIC_32;
            envP->coLi = 1;   // only Li0 is used in this demo
            envP->coHighways = 1;
            envP->coHdlcExDevs = 32;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'O') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == '2') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else if (*(argv[i]+2) == '6') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_6;
          } else if (*(argv[i]+2) == '7') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_7;
          } else if (*(argv[i]+2) == '8') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_8;
          } else if (*(argv[i]+2) == '9') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_9;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '0') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_10;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '1') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_11;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '2') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_12;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '3') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_13;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '4') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_14;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '5') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_15;
          } else if ( (*(argv[i]+2) == 'H') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_A_CLOCKS_ETCF;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i] + 1) == 'M') {
          // Li Mode
          if (*(argv[i] + 2) == 'E') {
            // E1
            envP->eT1E1Mode = OTX_T1E1_MODE_E1;
          } else if (*(argv[i] + 2) == 'T') {
            // T1
            envP->eT1E1Mode = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name (default=%s)\n", DEMO_DEFAULT_LOG_FILE);
  DemoLogf(" /R<fileName> : File name for decoded HDLC frames (default=%s)\n", DEMO_DEFAULT_FRAMES_FILE);
  DemoLogf(" /B<boardNo>  : First board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /C<boards>   : Number of boards (Default is %d)\n", 1);
  DemoLogf(" /T[T|4|P|S|G]: Board Type used for the demo, where\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  S = Use Sleipnir Board\n");
  DemoLogf("                  G = Use Gimle-16-PCI-Plus Board\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n\n");
  DemoLogf(" /5           : Use 56kbps HDLC channels (instead of default 64kpbs channels).\n");
  DemoLogf(" /S[N|S|A]    : SS7 FISU/LSSU filtering, where\n");
  DemoLogf("                  N = No filtering at all\n");
  DemoLogf("                  S = Filter all duplicate FISUs and LSSUs (default)\n");
  DemoLogf("                  A = Filter all FISUs and LSSUs\n");
  DemoLogf(" /K[F|A]      : HDLC Interframe Timefill Character, where\n");
  DemoLogf("                  F = Flags (0x7E)(default)\n");
  DemoLogf("                  A = All Ones (0xFF)\n");
  DemoLogf(" /N           : Decode HDLC on TS0\n");
  DemoLogf(" /I           : Use HDLC Bit Inversion (Default is to not invert the bits)\n");
  DemoLogf(" /O[I|x]        : Clock Source, where\n");
  DemoLogf("                  x = I Internal clock source (default) or\n");
  DemoLogf("                  x = 0 through 15 indicating with Li to recover the clock from or\n");
  DemoLogf("                  x = H to recover from H.100 A-clock (ETCF mode)\n");
  DemoLogf(" /A           : Front End Amplifier = ON (default is OFF)\n");
  DemoLogf(" /E           : Receive Equalizer = ON (longhaul) (default is OFF)\n");
  DemoLogf(" /2           : -20dB MonitorMode = ON (default is OFF)\n");
  DemoLogf(" /H           : Receive Input Threshold in mV (default is %d mV)\n", DEMO_DEFAULT_RX_THRESHOLD);
  DemoLogf(" /Y           : Termination Mode, where\n");
  DemoLogf("                  1 = 100/120 ohm balanced\n");
  DemoLogf("                  7 = 75 ohm unbalanced\n");
  DemoLogf("                  Z = High-impedance\n");
  DemoLogf("                  S = 75 ohm balanced\n");
  DemoLogf(" /P           : Tri-state transmit pins (default is drive transmit pins)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
  DemoLogf(" /V           : Non-Verbose Logging (Default is verbose).\n");
  DemoLogf(" /W<dev>[,:<dev>..]  : Enable specific list of RX devices only\n");
  DemoLogf("                List device numbers separated by \',\'\n");
  DemoLogf(" /F[E|S|9]    : T1 Frame Format, where\n");
  DemoLogf("                  E = ESF (default)\n");
  DemoLogf("                  4 = F4\n");
  DemoLogf("                  S = SF [D3/D4] (F12)\n");
  DemoLogf("                  9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]      : T1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]    : E1 Frame Format, where\n");
  DemoLogf("                  D = Doubleframe\n");
  DemoLogf("                  C = CRC4 Multiframe (default)\n");
  DemoLogf("                  G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]      : E1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  H = HDB3 (default)\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char* formatStr, ...)
{
  va_list argList;

  //EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile, formatStr, argList);
  }

  va_end(argList);

  //LeaveCriticalSection(&g_demoLogfCritSec);
}


/*******************************************************************************

  @func

  prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf2(char* formatStr, ...)
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile, formatStr, argList);
  }

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);

  // close file when it exists
  if (g_pDemoFramesFile != NULL) {
    fclose(g_pDemoFramesFile);
  }
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char* argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;
  OTX_INT32 nBrd;

  OsaConsoleInit();

  printf("HdlcExRxManyTs, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for the OtxHdlcEx Library.\n");
  printf("Multiple 64kpbs HDLC receivers per E1/T1 span.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

  if (g_demoEnv.framesFileName[0]) {
    if ((g_pDemoFramesFile = fopen(g_demoEnv.framesFileName, "wt")) == NULL) {
      printf("Couldn't open HDLC Frame Log file %s\n", g_demoEnv.framesFileName);
      OsaConsoleExit();
      exit(-1);
    }
  }

  atexit(DemoAtExit);

  DemoLogf("Log file is:           %s\n", g_demoEnv.logFileName);
  DemoLogf("Frame file is:         %s\n", g_demoEnv.framesFileName);
  DemoLogf("Starting board number: %d\n", g_demoEnv.firstBoardNo);
  DemoLogf("Number of boards:      %d\n", g_demoEnv.nBoards);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static" : "dynamic");
  }

  DemoLogf("Connecting OtxHdlcEx lib...\n");
  result = OtxHdlcExConnectLib();

  switch (OTX_RESULT_CODE(result)) {
   case OTX_RES_HANDLE:
     DemoLogf("Error: Could not connect to the OTX HdlcEx library.\n");
     OsaConsoleExit();
     OtxDrvReleaseLib();
     exit(1);
     break;
   default:
     DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
     break;
  }

  // Init all lobal objects 
  DemoInitializeGlobalData();

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  for (nBrd = 0; nBrd < g_demoEnv.nBoards; nBrd++) {
    
    // Connect (get handles) to the physical devices on the OTX board
    DemoLogf("%d: DemoCreatePhysicalDevices\n", nBrd);
    DemoCreatePhysicalDevices(nBrd);
    
    // Set clock source for the board
    {
      char szClockSource[64];
      
      switch (g_demoEnv.eClkSource) {
      case OTX_CLOCK_SOURCE_INTERNAL:
        strcpy(szClockSource, "Internal");
        break;
      case OTX_CLOCK_SOURCE_LOCAL_0:
      case OTX_CLOCK_SOURCE_LOCAL_1:
      case OTX_CLOCK_SOURCE_LOCAL_2:
      case OTX_CLOCK_SOURCE_LOCAL_3:
      case OTX_CLOCK_SOURCE_LOCAL_4:
      case OTX_CLOCK_SOURCE_LOCAL_5:
      case OTX_CLOCK_SOURCE_LOCAL_6:
      case OTX_CLOCK_SOURCE_LOCAL_7:
      case OTX_CLOCK_SOURCE_LOCAL_8:
      case OTX_CLOCK_SOURCE_LOCAL_9:
      case OTX_CLOCK_SOURCE_LOCAL_10:
      case OTX_CLOCK_SOURCE_LOCAL_11:
      case OTX_CLOCK_SOURCE_LOCAL_12:
      case OTX_CLOCK_SOURCE_LOCAL_13:
      case OTX_CLOCK_SOURCE_LOCAL_14:
      case OTX_CLOCK_SOURCE_LOCAL_15:
        sprintf(szClockSource, "Line Interface %d", (int)g_demoEnv.eClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
        break;
      case OTX_CLOCK_SOURCE_A_CLOCKS_ETCF:
        sprintf(szClockSource, "H.100 A-Clocks (ETCF mode)");
        break;
      default:
        strcpy(szClockSource, "Unknown");
        break;
      }
      
      DemoLogf("%d: Clock Source: %s\n", nBrd, szClockSource);
      
      result = OtxBrdSetClocks(g_hPhysDevs[nBrd].hBoardController, OTX_CLOCK_MODE_H100_ETCF, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs[nBrd].hBoardController, result);
    }
    
    
    
    // Configure E1/T1 devices
    DemoLogf("%d: Configuring E1/T1 devices...\n", nBrd);
    DemoConfigureE1T1Devices(nBrd);
    
    // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
    DemoLogf("%d: DemoEnablePhysicalDevices\n", nBrd);
    DemoEnablePhysicalDevices(nBrd);
    
    // Create and get handles to the Pipe devices needed by this application
    DemoLogf("%d: DemoConnectDevices\n", nBrd);
    DemoConnectDevices(nBrd);
    
    // Create and get handles to the Logical devices needed by this application
    DemoLogf("%d: Creating logical devices...\n", nBrd);
    DemoCreateLogicalDevices(nBrd);
    
    // Initialized the Connected Logical Devices
    DemoLogf("%d: Initializing logical devices...\n", nBrd);
    DemoInitializeLogicalDevices(nBrd);
    
    // Create and get handles to the Pipe devices needed by this application
    DemoLogf("%d: Connecting logical devices...\n", nBrd);
    DemoConnectLogicalDevices(nBrd);
    
    // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
    DemoLogf("%d: Enabling logical devices...\n", nBrd);
    DemoEnableLogicalDevices(nBrd);
    
    DemoStartBurstDevices(nBrd);

  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  OtxDrvGetCurrentTime(&g_nTimeStampForBytePerSecCalculation);

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // abort threads
  DemoLogf("\nExiting...\n"); 
  
  {
    OTX_INT32 i;


    g_bStatsThreadActive = OTX_FALSE;

    for (nBrd = 0; nBrd < g_demoEnv.nBoards; nBrd++) {
      for (i = 0; i < g_demoEnv.coHighways; i++) {
        g_hLogicDevs[nBrd].mBurstReceiver[i].bActive = OTX_FALSE;
      }
    }

    OsaThreadSleep(1500);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();

#if 1   // this is NOT needed w/ 1.05 version of OtxHdlcEx library
  OtxDrvReleaseLib();
#endif

  OtxHdlcExReleaseLib();

  return 0;
}

